#!/usr/bin/env python
# coding: utf-8

# Copyright (c) Saga Inc.
# Distributed under the terms of the GPL License.
"""
Contains helpful utility functions for saving and reading
in analyses.
"""

from mitosheet.step import Step
import os
import json
from typing import Any, Dict, List, Optional
from mitosheet._version import __version__
from mitosheet.telemetry.telemetry_utils import log
from mitosheet.types import StepsManagerType
from mitosheet.utils import NpEncoder
from mitosheet.save_paths import MITO_FOLDER

# The current version of the saved Mito analysis
# where we save all the analyses for this version
SAVED_ANALYSIS_FOLDER = os.path.join(MITO_FOLDER, 'saved_analyses')


def get_analysis_exists(analysis_name: Optional[str]) -> bool:
    """
    Given an analysis_name, returns if the saved analysis in
    ~/.mito/{analysis_name}.json exists
    """

    if analysis_name is None:
        return False

    analysis_path = f'{SAVED_ANALYSIS_FOLDER}/{analysis_name}.json'
    return os.path.exists(analysis_path)

def read_analysis(analysis_name: str) -> Optional[Dict[str, Any]]:
    """
    Given an analysis_name, reads the saved analysis in
    ~/.mito/{analysis_name}.json and returns a JSON object
    representing it.
    """

    analysis_path = f'{SAVED_ANALYSIS_FOLDER}/{analysis_name}.json'

    if not os.path.exists(analysis_path):
        return None

    with open(analysis_path) as f:
        try:
            # We try and read the file as JSON
            return json.load(f)
        except: 
            return None

def read_and_upgrade_analysis(analysis_name: str) -> Optional[Dict[str, Any]]:
    """
    Given an analysis_name, reads the saved analysis in
    ~/.mito/{analysis_name}.json, does it's best to upgrade it to the current
    saved version, and then returns it.
    """
    from mitosheet.saved_analyses import upgrade_saved_analysis_to_current_version
    old_analysis = read_analysis(analysis_name)
    return upgrade_saved_analysis_to_current_version(old_analysis)

def _get_all_analysis_filenames():
    """
    Returns the names of the files in the SAVED_ANALYSIS_FOLDER
    """
    if not os.path.exists(SAVED_ANALYSIS_FOLDER):
        return []

    file_names = set([
        f for f in os.listdir(SAVED_ANALYSIS_FOLDER) 
        if os.path.isfile(os.path.join(SAVED_ANALYSIS_FOLDER, f))
    ])

    return file_names

def _delete_analyses(analysis_filenames):
    """
    For bulk deleting analysis with file names. 
    """
    for filename in analysis_filenames:
        os.remove(os.path.join(SAVED_ANALYSIS_FOLDER, filename))

def delete_saved_analysis(analysis_name):
    """
    Deletes a saved analysis. saved_analysis_file_name must end in .json

    Throws an error if analysis_name does not exist.
    """

    analysis = read_analysis(analysis_name)

    # If the analysis name exists, delete it
    if analysis is not None:
        os.remove(os.path.join(SAVED_ANALYSIS_FOLDER, analysis_name + '.json'))
    else:
        raise Exception(f'Cannot delete {analysis_name} as it does not exist')


def rename_saved_analysis(old_analysis_name, new_analysis_name):
    """
    Renames a saved analysis from old_analysis_name to new_analysis_name. 

    Throws an error if old_analysis_name does not exist, or new_analysis_name
    exists.
    """
    old_analysis = read_analysis(old_analysis_name)
    new_analysis = read_analysis(new_analysis_name)

    # If the old_analysis_file_name exists, and new_analysis_file_name does not exist
    if old_analysis is not None and new_analysis is None:
        full_old_analysis_name = os.path.join(SAVED_ANALYSIS_FOLDER, old_analysis_name + '.json')
        full_new_analysis_name = os.path.join(SAVED_ANALYSIS_FOLDER, new_analysis_name + '.json')
        os.rename(full_old_analysis_name, full_new_analysis_name)
    else:
        raise Exception(f'Invalid rename, with old and new analysis are {old_analysis_name} and {new_analysis_name}')


def write_saved_analysis(analysis_path: str, steps_data: List[Dict[str, Any]], version: str=__version__) -> None:
    with open(analysis_path, 'w+') as f:
        saved_analysis = {
            'version': version,
            'steps_data': steps_data
        }

        f.write(json.dumps(saved_analysis, cls=NpEncoder))


def make_steps_json_obj(
        steps: List[Step]
    ) -> List[Dict[str, Any]]:
    """
    Given a steps dictonary from a steps_manager, puts the steps
    into a format that can be saved to a json file. 

    The file format is currently: 
    {
        'version': '...',
        'steps_data': [
            {
                'step_version': 1
                'step_type': 'add_column_event',
                'params': {
                    ...
                }
            }
        ]
    }

    Notably, does not return any skipped steps, which is necessary
    because we don't save the step id, so then we cannot detect
    which should be skipped properly
    """
    from mitosheet.steps_manager import get_step_indexes_to_skip

    steps_json_obj = []

    skipped_step_indexes = get_step_indexes_to_skip(steps)

    for step_index, step in enumerate(steps):
        # Skip the initialize step
        if step.step_type == 'initialize':
            continue

        # Skip the skipped steps
        if step_index in skipped_step_indexes:
            continue

        # Save the step type
        step_summary = {
            'step_version': step.step_performer.step_version(),
            'step_type': step.step_type,
            'params': step.params
        }

        steps_json_obj.append(step_summary)                

    return steps_json_obj

def write_analysis(steps_manager: StepsManagerType, analysis_name: Optional[str]=None) -> None:
    """
    Writes the analysis saved in steps_manager to
    ~/.mito/{analysis_name}. If analysis_name is none, gets the temporary
    name from the steps_manager.

    Note that a step container may contain invalid steps/out of
    date steps, but we save them all, as they will play back validly
    as they were valid when they were added.
    """

    if not os.path.exists(MITO_FOLDER):
        os.mkdir(MITO_FOLDER)

    if not os.path.exists(SAVED_ANALYSIS_FOLDER):
        os.mkdir(SAVED_ANALYSIS_FOLDER)

    if analysis_name is None:
        analysis_name = steps_manager.analysis_name

    analysis_path = f'{SAVED_ANALYSIS_FOLDER}/{analysis_name}.json'
    steps = make_steps_json_obj(steps_manager.steps_including_skipped)

    # Actually write the file
    write_saved_analysis(analysis_path, steps)


def register_analysis(analysis_name):
    """
    A helper function that is used in the frontend
    to save which analysis is being run (so it can be replayed)
    """
    log('ran_generated_code', {'analysis_name': analysis_name}) 


#!/usr/bin/env python
# coding: utf-8

# Copyright (c) Saga Inc.
# Distributed under the terms of the GPL License.
"""
Rolls back/forward to a specific step, by index
"""

from mitosheet.types import StepsManagerType


CHECKOUT_STEP_BY_IDX_UPDATE_EVENT = 'checkout_step_by_idx_update'
CHECKOUT_STEP_BY_IDX_UPDATE_PARAMS = [
    'step_idx'
]

def execute_checkout_step_by_idx_update(
        steps_manager: StepsManagerType,
        step_idx: int
    ) -> None:
    """
    Checks out a specific step by index
    """
    # If the new step index is -1, go to the end of the list!
    if step_idx == -1:
        step_idx = len(steps_manager.steps_including_skipped) - 1

    steps_manager.curr_step_idx = step_idx

CHECKOUT_STEP_BY_IDX_UPDATE = {
    'event_type': CHECKOUT_STEP_BY_IDX_UPDATE_EVENT,
    'params': CHECKOUT_STEP_BY_IDX_UPDATE_PARAMS,
    'execute': execute_checkout_step_by_idx_update
}


#!/usr/bin/env python
# coding: utf-8

# Copyright (c) Saga Inc.
# Distributed under the terms of the GPL License.
"""
Contains functions that are useful for determining the state of the
current user.
"""
import getpass
import hashlib
import os
from datetime import datetime
import sys
from typing import Optional

import pandas as pd
from mitosheet._version import __version__, package_name
from mitosheet.user.db import get_user_field
from mitosheet.user.schemas import (UJ_MITOSHEET_LAST_UPGRADED_DATE,
                                    UJ_MITOSHEET_PRO)

try:
    import mitosheet_helper_pro
    MITOSHEET_HELPER_PRO = True
except ImportError:
    MITOSHEET_HELPER_PRO = False


def is_running_test() -> bool:
    """
    A helper function that quickly returns if the current code is running 
    inside of a test, which is useful for making sure we don't generate 
    tons of logs.
    """
    # Pytest injects PYTEST_CURRENT_TEST into the current environment when running
    running_pytests = "PYTEST_CURRENT_TEST" in os.environ
    # Github injects CI into the environment when running
    running_ci = 'CI' in os.environ and os.environ['CI'] is not None

    return running_pytests or running_ci

def is_on_kuberentes_mito() -> bool:
    """
    Returns True if the user is on Kuberentes Mito, on staging or on app
    """
    user = getpass.getuser()
    return user == 'jovyan'

def is_pro() -> bool:
    """
    Helper function for returning if this is a
    pro deployment of mito
    """
    is_pro = get_user_field(UJ_MITOSHEET_PRO)

    # This package overides the user.json
    if MITOSHEET_HELPER_PRO:
        return MITOSHEET_HELPER_PRO

    # If the current package is mitosheet-private, we activate Pro
    if package_name == 'mitosheet-private':
        return True

    return is_pro if is_pro is not None else False


def is_local_deployment() -> bool:
    """
    Helper function for figuring out if this a local deployment or a
    Mito server deployment
    """
    return not is_on_kuberentes_mito()  


def should_upgrade_mitosheet() -> bool:
    """
    A helper function that calculates if a user should upgrade, which does so by 
    checking if the user has upgraded in the past 21 days (3 weeks), since this is
    about how often we release big features.

    Always returns false if:
    - it is not a local installation, for obvious reasons.
    - if it has an admin package installed, as this is managed by an admin

    NOTE: if the user clicks the upgrade button in the app, then we change the upgraded 
    date to this date, so that the user doesn't get a bunch of annoying popups. This just
    pushes back when they are annoyed to upgrade!
    """
    if not is_local_deployment():
        return False
    
    from mitosheet.telemetry.telemetry_utils import MITOSHEET_HELPER_PRIVATE
    if MITOSHEET_HELPER_PRO or MITOSHEET_HELPER_PRIVATE:
        return False

    # If it's mitosheet-private, then we don't give them the upgrade prompts
    if package_name == 'mitosheet-private':
        return False
    
    last_upgraded_date_stored = get_user_field(UJ_MITOSHEET_LAST_UPGRADED_DATE)
    if last_upgraded_date_stored is None:
        return False

    mitosheet_last_upgraded_date = datetime.strptime(last_upgraded_date_stored, '%Y-%m-%d')
    return (datetime.now() - mitosheet_last_upgraded_date).days > 21

def get_pandas_version() -> str:
    """
    Returns the pandas version
    """
    return pd.__version__


def get_python_version() -> int:
    """
    Returns the Python version
    """
    return sys.version_info.minor


def check_pro_acccess_code(access_code: Optional[str]) -> bool:
    """Checks if the passed access code is correct, by hashing it and comparing to the hashed value"""
    return access_code is not None and hashlib.sha256(access_code.encode()).hexdigest() == '761a24dea594a8eafe698acfebb77de90bf0826c9400a2543500ee98929ea132'


#!/usr/bin/env python
# coding: utf-8

# Copyright (c) Saga Inc.
# Distributed under the terms of the GPL License.
from typing import (Any, Collection, Dict, List, Optional,
                    Tuple, Union)

import pandas as pd
from mitosheet.code_chunks.step_performers.import_steps.simple_import_code_chunk import generate_read_csv_code
from mitosheet.errors import get_recent_traceback_as_list
from mitosheet.telemetry.telemetry_utils import log
from mitosheet.preprocessing.preprocess_step_performer import \
    PreprocessStepPerformer
from mitosheet.step_performers.import_steps.simple_import import (
    get_valid_dataframe_names, read_csv_get_delimiter_and_encoding)
from mitosheet.types import StepsManagerType


class ReadFilePathsPreprocessStepPerformer(PreprocessStepPerformer):
    """
    This preprocessor reads in any arguments that are
    strings, treats them as file paths, and attempts
    to read them in a dataframes.
    """

    @classmethod
    def preprocess_step_version(cls) -> int:
        return 1

    @classmethod
    def preprocess_step_type(cls) -> str:
        return 'read_file_paths'

    @classmethod
    def execute(cls, args: Collection[Any]) -> Tuple[List[Any], Dict[str, Any]]:
        df_args: List[pd.DataFrame] = []
        delimeters: List[Optional[str]] = []
        encodings: List[Optional[str]] = []
        for arg in args:
            if isinstance(arg, pd.DataFrame):
                df_args.append(arg)
                delimeters.append(None)
                encodings.append(None)
            elif isinstance(arg, str):
                # If it is a string, we try and read it in as a dataframe
                try:
                    # We use the simple import 
                    df, delimeter, encoding = read_csv_get_delimiter_and_encoding(arg)

                    df_args.append(
                        df
                    )

                    delimeters.append(delimeter)
                    encodings.append(encoding)
                except:
                    # If this pd.read_csv fails, then we report this error to the user
                    # as a failed mitosheet call
                    error_message = f'Invalid argument passed to sheet: {arg}. This path could not be read with a pd.read_csv call. Please pass in the parsed dataframe directly.'
                    log('mitosheet_sheet_call_failed', failed=True)
                    raise ValueError(error_message)
            else:
                error_message = f'Invalid argument passed to sheet: {arg}. Please pass all dataframes or paths to CSV files.'
                log('mitosheet_sheet_call_failed', {'error': error_message}, failed=True)
                raise ValueError(error_message)
                
        return df_args, {
            'delimeters': delimeters,
            'encodings': encodings,
        }

    @classmethod
    def transpile(cls, steps_manager: StepsManagerType, execution_data: Optional[Dict[str, Any]]) -> List[str]:
        """
        Transpiles the reading in of passed file paths to dataframe names, 
        with a simple pd.read_csv call.
        """
        code = []

        # First, we get all the string arguments passed to the sheet call
        str_args = get_string_args(steps_manager.original_args)
        
        # Then, we turn these into dataframe names.
        # NOTE: there is potentially a bug if a user passes in a dataframe
        # with the same name as the the result of the return of this function,
        # but we ignore this for now. E.g. mitosheet.sheet('df1.csv', 'df1_csv') 
        # will cause only variable to exist.
        df_names = get_valid_dataframe_names([], str_args)

        delimeters = execution_data['delimeters'] if execution_data is not None else [None for _ in range(len(df_names))]
        encodings = execution_data['encodings'] if execution_data is not None else [None for _ in range(len(df_names))]

        num_strs = 0
        for arg_index, arg in enumerate(steps_manager.original_args):
            if isinstance(arg, str):
                df_name = df_names[num_strs]
                num_strs += 1

                read_csv_code = generate_read_csv_code(
                    arg, df_name, delimeters[arg_index], encodings[arg_index], None, None, None
                )

                code.append(
                    read_csv_code
                )

        if len(code) > 0:
            code.insert(0, '# Read in filepaths as dataframes')
                
        return code


def get_string_args(args: Collection[Union[pd.DataFrame, str]]) -> List[str]:
    return [arg for arg in args if isinstance(arg, str)]



#!/usr/bin/env python
# coding: utf-8

# Copyright (c) Saga Inc.
# Distributed under the terms of the GPL License.

import json
from typing import Any, Dict
from mitosheet.types import StepsManagerType


def get_render_count(params: Dict[str, Any], steps_manager: StepsManagerType) -> int:
    return steps_manager.render_count



#!/usr/bin/env python
# coding: utf-8

# Copyright (c) Saga Inc.
# Distributed under the terms of the GPL License.
"""
Utilities to help with type functions
"""

from typing import Any, List, Optional, Tuple, Union
import pandas as pd
import numpy as np

from mitosheet.sheet_functions.sheet_function_utils import is_series_of_constant

# A series of helper functions that help you figure
# out which dtype we're dealing with. NOTE: since some
# of these types can be different varieties (e.g. int can be int64, uint64)
# we try to check for them with simple expressions
# NOTE: these should be identical to the TS utilities in dtypes.tsx

def is_bool_dtype(dtype: str) -> bool:
    return 'bool' == dtype

def is_int_dtype(dtype: str) -> bool:
    return 'int' in dtype

def is_float_dtype(dtype: str) -> bool:
    return 'float' in dtype

def is_string_dtype(dtype: str) -> bool:
    return dtype == 'object' or dtype == 'str' or dtype == 'string'

def is_datetime_dtype(dtype: str) -> bool:
    # NOTE: this should handle all different datetime columns, no matter
    # the timezone, as it checks for string inclusion
    return 'datetime' in dtype

def is_timedelta_dtype(dtype: str) -> bool:
    return 'timedelta' in dtype

def is_number_dtype(dtype: str) -> bool:
    return is_int_dtype(dtype) or is_float_dtype(dtype)

def is_none_type(value: Union[str, None]) -> bool:
    """
    Helper function for determining if a value should be treated as None
    """
    return True if value is None or str(value).lower() in ['nan', 'nat'] else False

def get_float_dt_td_columns(df: pd.DataFrame) -> Tuple[List[Any], List[Any], List[Any]]:
    float_columns, date_columns, timedelta_columns = [], [], []
    for column_header in df.columns:
        dtype = str(df[column_header].dtype)
        # NOTE: these functions are called frequently, so we put them in 
        # the order they are most likely to be true in, so we can short out
        if is_float_dtype(dtype):
            float_columns.append(column_header)
        elif is_datetime_dtype(dtype):
            date_columns.append(column_header)
        elif is_timedelta_dtype(dtype):
            timedelta_columns.append(column_header)

    return float_columns, date_columns, timedelta_columns


def get_nan_indexes_metadata(*argv: pd.Series) -> Tuple[pd.Index, pd.Index]: 
    """
    Given a list of series, this function returns data that is helpful
    in figuring out which of the rows of these series have a NaN in them.

    The data returned allows you to remove the nan values from the series,
    so that they can then be easily used by sheet functions, but also to
    easily add these NaN values back to the series in the correct location.

    It does so by returning a tuple of the original_index, non_nan_index 
    """
    nan_index_set = set()
    non_nan_index_set = set()
    original_index = None

    for arg in argv:
        if isinstance(arg, pd.Series): 
            nan_index_set.update(arg[arg.isnull()].index)
            non_nan_index_set.update(arg[~arg.isnull()].index)

            # Update the original indexes to the first element, if we haven't
            # selected one yet
            if original_index is None:
                original_index = arg.index
            elif not is_series_of_constant(arg):
                original_index = arg.index

    # Remove any index that was nan at any point
    non_nan_index_set.difference_update(nan_index_set)
    non_nan_index = pd.Index(list(non_nan_index_set))

    return original_index, non_nan_index 

def put_nan_indexes_back(series: pd.Series, original_index: pd.Index) -> pd.Series:
    """
    This function takes a series, as well as a set of the indexes
    that are are NaN, and inserts these indexes back into the series
    as NaN values.
    """
    return series.reindex(original_index)


def get_datetime_format(string_series: pd.Series) -> Optional[str]:
    """
    Given a series of datetime strings, guesses the most likely date format.
    """
    try:
        # Import log function here to avoid circular import
        from mitosheet.telemetry.telemetry_utils import log

        # If we can convert all non null inputs, then we assume we guessed correctly
        non_null_inputs = string_series[~string_series.isna()]

        # First try letting pandas guess the correct datetime
        converted = pd.to_datetime(non_null_inputs, errors='coerce', infer_datetime_format=True)
        if converted.isna().sum() == 0:
            return None

        # TODO: Add the most popular formats to here and check them first before 
        # trying all of the formats below for performance.

        # Then we try a bunch of other formats it could be
        sample_string_datetime = string_series[string_series.first_valid_index()]
        FORMATS = [
            '%m{s}%d{s}%Y', 
            '%d{s}%m{s}%Y', 
            '%Y{s}%d{s}%m', 
            '%Y{s}%m{s}%d', 
            '%m{s}%d{s}%Y %H:%M:%S', 
            '%d{s}%m{s}%Y  %H:%M:%S', 
            '%Y{s}%d{s}%m  %H:%M:%S', 
            '%Y{s}%m{s}%d  %H:%M:%S'
        ]
        SEPERATORS = ['/', '-', '.', ',', ':', ' ', '']

        for seperator in SEPERATORS:
            if seperator in sample_string_datetime:
                for _format in FORMATS:
                    format = _format.format(s=seperator)
                    if test_datetime_format(non_null_inputs, format):
                        return format  

        log('unable_to_determine_datetime_format_on_cast')
        return None
    except:
        log('unable_to_determine_datetime_format_on_cast')
        return None


def test_datetime_format(series: pd.Series, _format: str) -> bool:
    converted = pd.to_datetime(series, errors='coerce', format=_format)
    return converted.isna().sum() == 0
        

def get_million_identifier_in_string(string: str) -> Union[str, None]:
    """
    Given a string, returns the million identifier in it. 
    Returns '' if none exist. 
    """
    million_identifiers = ["Million", 'Mil', 'M', 'million', 'mil', 'm']
    # So that we return the biggest matching element
    million_identifiers = list(sorted(million_identifiers, key = len, reverse=True))

    for identifier in million_identifiers:
        if identifier in string:
            return identifier

    return None


def get_billion_identifier_in_string(string: str) -> Union[str, None]:
    """
    Given a string, returns the billion identifier in it. 
    Returns '' if none exist. 
    """

    billion_identifiers = ["Billion", 'Bil', 'B', 'billion', 'bil', 'b']
    # So that we return the biggest matching element
    billion_identifiers = list(sorted(billion_identifiers, key = len, reverse=True))

    for identifier in billion_identifiers:
        if identifier in string:
            return identifier

    return None


#!/usr/bin/env python
# coding: utf-8

# Copyright (c) Mito.
# Distributed under the terms of the Modified BSD License.
from copy import deepcopy
from typing import TYPE_CHECKING, Any, Dict, Collection, List, Optional, Tuple
import pandas as pd
from mitosheet.preprocessing.preprocess_step_performer import \
    PreprocessStepPerformer
from mitosheet.types import StepsManagerType


class CopyPreprocessStepPerformer(PreprocessStepPerformer):
    """
    This preprocessing step is responsible for making a copy of all of the
    passed arguments, so that dataframes aren't modified incorrectly.
    """

    @classmethod
    def preprocess_step_version(cls) -> int:
        return 1

    @classmethod
    def preprocess_step_type(cls) -> str:
        return 'copy'

    @classmethod
    def execute(cls, args: Collection[Any]) -> Tuple[List[Any], Optional[Dict[str, Any]]]:
        
        new_args = []
        for arg in args:
            if isinstance(arg, pd.DataFrame):
                # Do a pandas copy if it's a dataframe
                arg_copy = arg.copy(deep=True)
            else:
                # Simple deepcopy if it's a string
                arg_copy = deepcopy(arg)
            new_args.append(arg_copy)
        return new_args, None

    @classmethod
    def transpile(cls, steps_manager: StepsManagerType, execution_data: Optional[Dict[str, Any]]) -> List[str]:
        return []



#!/usr/bin/env python
# coding: utf-8

# Copyright (c) Saga Inc.
# Distributed under the terms of the GPL License.
"""
Contains all functions that can be used in a sheet that operate on strings.

All functions describe their behavior with a function documentation object
in the function docstring. Function documentation objects are described
in more detail in docs/README.md.
NOTE: This file is alphabetical order!
"""
import functools

import pandas as pd
from mitosheet.sheet_functions.sheet_function_utils import (
    fill_series_with_one_index, try_extend_series_to_index)
from mitosheet.sheet_functions.types.decorators import (
    cast_output, convert_arg_to_series_type, convert_args_to_series_type,
    filter_nans, handle_sheet_function_errors)


@handle_sheet_function_errors
@filter_nans
@convert_arg_to_series_type(
    0,
    'string',
    'error',
    'error'
)
def CLEAN(series: pd.Series) -> pd.Series:
    """
    {
        "function": "CLEAN",
        "description": "Returns the text with the non-printable ASCII characters removed.",
        "search_terms": ["clean", "trim", "remove"],
        "examples": [
            "CLEAN(A)"
        ],
        "syntax": "CLEAN(string)",
        "syntax_elements": [{
                "element": "string",
                "description": "The string or series whose non-printable characters are to be removed."
            }
        ]
    }
    """
    # TODO: this function is wacked out. It removes spaces (Excel's does not!)
    # Also, I'm not sure it makes any sense in the context of Python + pandas
    # reading in dataframes, given they handle all characters here.
    return series.apply(lambda x:''.join([i if 32 <= ord(i) < 126 else "" for i in x]))


@handle_sheet_function_errors
@filter_nans
@convert_args_to_series_type('string', 'skip', ('default', ''))
def CONCAT(*argv: pd.Series) -> pd.Series:
    """
    {
        "function": "CONCAT",
        "description": "Returns the passed strings and series appended together.",
        "search_terms": ["&", "concatenate", "append", "combine"],
        "examples": [
            "CONCAT('Bite', 'the bullet')",
            "CONCAT(A, B)"
        ],
        "syntax": "CONCAT(string1, [string2, ...])",
        "syntax_elements": [{
                "element": "string1",
                "description": "The first string or series."
            },
            {
                "element": "string2, ... [OPTIONAL]",
                "description": "Additional strings or series to append in sequence."
            }
        ]
    }
    """

    # We make sure all the series are the max length, so the concat has something at every index
    argv = fill_series_with_one_index(argv)

    return functools.reduce((lambda x, y: x + y), argv)


@handle_sheet_function_errors
@filter_nans
@convert_arg_to_series_type(
    0,
    'string',
    'error',
    'error'
)
@convert_arg_to_series_type(
    1,
    'string',
    'error',
    'error'
)
def FIND(series: pd.Series, substrings: pd.Series) -> pd.Series:
    """
    {
        "function": "FIND",
        "description": "Returns the position at which a string is first found within text, case-sensitive. Returns 0 if not found.",
        "search_terms": ["find", "search"],
        "examples": [
            "FIND(A, 'Jack')",
            "FIND('Ben has a friend Jack', 'Jack')"
        ],
        "syntax": "FIND(text_to_search, search_for)",
        "syntax_elements": [{
                "element": "text_to_search",
                "description": "The text or series to search for the first occurrence of search_for."
            },
            {
                "element": "search_for",
                "description": "The string to look for within text_to_search."
            }
        ]
    }
    """

    # If there aren't enough substrings, we fill it to the end
    substrings = try_extend_series_to_index(substrings, series.index)

    # Then, we find each substring
    return pd.Series(
        [string.find(substring) + 1 for string, substring in zip(series, substrings)], 
        index=series.index
    )


@handle_sheet_function_errors
@cast_output('first_input_type')
@filter_nans
@convert_arg_to_series_type(
    0,
    'string',
    'error',
    'error'
)
@convert_arg_to_series_type(
    1,
    'string',
    'error',
    ('default', 1),
    optional=True
)
def LEFT(series: pd.Series, num_chars: pd.Series=None) -> pd.Series:
    """
    {
        "function": "LEFT",
        "description": "Returns a substring from the beginning of a specified string.",
        "search_terms": ["left"],
        "examples": [
            "LEFT(A, 2)",
            "LEFT('The first character!')"
        ],
        "syntax": "LEFT(string, [number_of_characters])",
        "syntax_elements": [{
                "element": "string",
                "description": "The string or series from which the left portion will be returned."
            },
            {
                "element": "number_of_characters [OPTIONAL, 1 by default]",
                "description": "The number of characters to return from the start of string."
            }
        ]
    }
    """
    if num_chars is None:
        num_chars = pd.Series(data=[1] * series.size)

    # If there aren't enough char splits, we fill it to the end
    num_chars = try_extend_series_to_index(num_chars, series.index)
    # And then slice the string on the left
    return pd.Series(
        [string[:int(num_char)] for string, num_char in zip(series, num_chars)],
        index=series.index
    )


@handle_sheet_function_errors
@filter_nans
@convert_arg_to_series_type(
    0,
    'string',
    'error',
    'error'
)
def LEN(series: pd.Series) -> pd.Series:
    """
    {
        "function": "LEN",
        "description": "Returns the length of a string.",
        "search_terms": ["length", "size"],
        "examples": [
            "LEN(A)",
            "LEN('This is 21 characters')"
        ],
        "syntax": "LEN(string)",
        "syntax_elements": [{
                "element": "string",
                "description": "The string or series whose length will be returned."
            }
        ]
    }
    """
    return series.str.len()


@handle_sheet_function_errors
@cast_output('first_input_type')
@filter_nans
@convert_arg_to_series_type(
    0,
    'string',
    'error',
    'error'
)
def LOWER(series: pd.Series) -> pd.Series:
    """
    {
        "function": "LOWER",
        "description": "Converts a given string to lowercase.",
        "search_terms": ["lowercase", "uppercase"],
        "examples": [
            "=LOWER('ABC')",
            "=LOWER(A)",
            "=LOWER('Nate Rush')"
        ],
        "syntax": "LOWER(string)",
        "syntax_elements": [{
                "element": "string",
                "description": "The string or series to convert to lowercase."
            }
        ]
    }
    """
    return series.str.lower()


@handle_sheet_function_errors
@cast_output('first_input_type')
@filter_nans
@convert_arg_to_series_type(
    0,
    'string',
    'error',
    'error'
)
@convert_arg_to_series_type(
    1,
    'int',
    'error',
    ('default', 1)
)
@convert_arg_to_series_type(
    2,
    'int',
    'error',
    ('default', 1)
)
def MID(series: pd.Series, start_loc: pd.Series, num_chars: pd.Series) -> pd.Series:
    """
    {
        "function": "MID",
        "description": "Returns a segment of a string.",
        "search_terms": ["middle"],
        "examples": [
            "MID(A, 2, 2)",
            "MID('Some middle characters!', 3, 4)"
        ],
        "syntax": "MID(string, starting_at, extract_length)",
        "syntax_elements": [{
                "element": "string",
                "description": "The string or series to extract the segment from."
            },
            {
                "element": "starting_at",
                "description": "The index from the left of string from which to begin extracting."
            },
            {
                "element": "extract_length",
                "description": "The length of the segment to extract."
            }
        ]
    }
    """
    # If there aren't enough char splits, we fill it to the end
    start_loc = try_extend_series_to_index(start_loc, series.index)
    num_chars = try_extend_series_to_index(num_chars, series.index)
    # And then slice the string on the left
    return pd.Series(
        [
            string[start - 1: start - 1 + int(num_char)] for string, start, num_char 
            in zip(series, start_loc, num_chars)
        ],
        index=series.index
    ) 


@handle_sheet_function_errors
@cast_output('first_input_type')
@filter_nans
@convert_arg_to_series_type(
    0,
    'string',
    'error',
    'error'
)
def PROPER(series: pd.Series) -> pd.Series:
    """
    {
        "function": "PROPER",
        "description": "Capitalizes the first letter of each word in a specified string.",
        "search_terms": ["proper", "capitalize"],
        "examples": [
            "=PROPER('nate nush')",
            "=PROPER(A)"
        ],
        "syntax": "PROPER(string)",
        "syntax_elements": [{
                "element": "string",
                "description": "The value or series to convert to convert to proper case."
            }
        ]
    }
    """
    return series.str.title()


@handle_sheet_function_errors
@cast_output('first_input_type')
@filter_nans
@convert_arg_to_series_type(
    0,
    'string',
    'error',
    'error'
)
@convert_arg_to_series_type(
    1,
    'int',
    'error',
    ('default', 1),
    optional=True
)
def RIGHT(series: pd.Series, num_chars: pd.Series=None) -> pd.Series:
    """
    {
        "function": "RIGHT",
        "description": "Returns a substring from the beginning of a specified string.",
        "search_terms": [],
        "examples": [
            "RIGHT(A, 2)",
            "RIGHT('The last character!')"
        ],
        "syntax": "RIGHT(string, [number_of_characters])",
        "syntax_elements": [{
                "element": "string",
                "description": "The string or series from which the right portion will be returned."
            },
            {
                "element": "number_of_characters [OPTIONAL, 1 by default]",
                "description": "The number of characters to return from the end of string."
            }
        ]
    }
    """
    if num_chars is None:
        num_chars = pd.Series(data=[1] * series.size)

    # If there aren't enough char splits, we fill it to the end
    num_chars = try_extend_series_to_index(num_chars, series.index)
    # And then slice the string on the left
    return pd.Series(
        [string[-int(num_char):] if num_char > 0 else '' for string, num_char in zip(series, num_chars)],
        index=series.index
    )


@handle_sheet_function_errors
@cast_output('first_input_type')
@filter_nans
@convert_arg_to_series_type(
    0,
    'string',
    'error',
    'error'
)
@convert_arg_to_series_type(
    1,
    'string',
    'error',
    'error'
)
@convert_arg_to_series_type(
    2,
    'string',
    'error',
    'error'
)
@convert_arg_to_series_type(
    3,
    'int',
    'error',
    ('default', -1),
    optional=True
)
def SUBSTITUTE(series: pd.Series, old_text: pd.Series, new_text: pd.Series, count: pd.Series=None) -> pd.Series:
    """
    {
        "function": "SUBSTITUTE",
        "description": "Replaces existing text with new text in a string.",
        "search_terms": ["replace", "find and replace"],
        "examples": [
            "SUBSTITUTE('Better great than never', 'great', 'late')",
            "SUBSTITUTE(A, 'dog', 'cat')"
        ],
        "syntax": "SUBSTITUTE(text_to_search, search_for, replace_with, [count])",
        "syntax_elements": [{
                "element": "text_to_search",
                "description": "The text within which to search and replace."
            },
            {
                "element": "search_for",
                "description": " The string to search for within text_to_search."
            },
            {
                "element": "replace_with",
                "description": "The string that will replace search_for."
            },
            {
                "element": "count",
                "description": "The number of times to perform the substitute. Default is all."
            }
        ]
    }
    """
    old_text = try_extend_series_to_index(old_text, series.index)
    new_text = try_extend_series_to_index(new_text, series.index)
    if count is None:
        count = try_extend_series_to_index(pd.Series(data=[-1]), series.index)

    return pd.Series(
        [string.replace(old, new, c) for string, old, new, c in zip(series, old_text, new_text, count)],
        index=series.index
    )

@handle_sheet_function_errors
@filter_nans
@convert_arg_to_series_type(
    0,
    'string',
    'error',
    'error'
)
def TEXT(series: pd.Series) -> pd.Series:
    """
    {
        "function": "TEXT",
        "description": "Turns the passed series into a string.",
        "search_terms": ["string", "dtype"],
        "examples": [
            "=TEXT(Product_Number)",
            "=TEXT(Start_Date)"
        ],
        "syntax": "TEXT(series)",
        "syntax_elements": [{
                "element": "series",
                "description": "The series to convert to a string."
            }
        ]
    }
    """
    # NOTE: we rely on the as type casting!
    return series


@handle_sheet_function_errors
@cast_output('first_input_type')
@filter_nans
@convert_arg_to_series_type(
    0,
    'string',
    'error',
    'error'
)
def TRIM(series: pd.Series) -> pd.Series:
    """
    {
        "function": "TRIM",
        "description": "Returns a string with the leading and trailing whitespace removed.",
        "search_terms": ["trim", "whitespace", "spaces"],
        "examples": [
            "=TRIM('  ABC')",
            "=TRIM('  ABC  ')",
            "=TRIM(A)"
        ],
        "syntax": "TRIM(string)",
        "syntax_elements": [{
                "element": "string",
                "description": "The value or series to remove the leading and trailing whitespace from."
            }
        ]
    }
    """
    return series.apply(lambda x: x.strip())


@handle_sheet_function_errors
@cast_output('first_input_type')
@filter_nans
@convert_arg_to_series_type(
    0,
    'string',
    'error',
    'error'
)
def UPPER(series: pd.Series) -> pd.Series:
    """
    {
        "function": "UPPER",
        "description": "Converts a given string to uppercase.",
        "search_terms": ["uppercase", "capitalize"],
        "examples": [
            "=UPPER('abc')",
            "=UPPER(A)",
            "=UPPER('Nate Rush')"
        ],
        "syntax": "UPPER(string)",
        "syntax_elements": [{
                "element": "string",
                "description": "The string or series to convert to uppercase."
            }
        ]
    }
    """

    return series.str.upper()


# TODO: we should see if we can list these automatically!
STRING_FUNCTIONS = {
    'CLEAN': CLEAN,
    'CONCAT': CONCAT,
    'FIND': FIND,
    'LEFT': LEFT,
    'LEN': LEN,
    'LOWER': LOWER,
    'MID': MID,
    'PROPER': PROPER,
    'RIGHT': RIGHT,
    'SUBSTITUTE': SUBSTITUTE,
    'TEXT': TEXT,
    'TRIM': TRIM,
    'UPPER': UPPER,
}



#!/usr/bin/env python
# coding: utf-8

# Copyright (c) Saga Inc.
# Distributed under the terms of the GPL License.
"""
See mito/mitosheet/steps/README.md for more information about 
how to add a step!
"""

from typing import Dict, List, Type
from mitosheet.step_performers.concat import ConcatStepPerformer
from mitosheet.step_performers.drop_duplicates import DropDuplicatesStepPerformer
from mitosheet.step_performers.import_steps.excel_import import ExcelImportStepPerformer
from mitosheet.step_performers.step_performer import StepPerformer
from mitosheet.step_performers.pivot import PivotStepPerformer
from mitosheet.step_performers.filter import FilterStepPerformer
from mitosheet.step_performers.sort import SortStepPerformer
from mitosheet.step_performers.set_cell_value import SetCellValueStepPerformer
from mitosheet.step_performers.column_steps.reorder_column import ReorderColumnStepPerformer
from mitosheet.step_performers.column_steps.add_column import AddColumnStepPerformer
from mitosheet.step_performers.column_steps.set_column_formula import SetColumnFormulaStepPerformer
from mitosheet.step_performers.merge import MergeStepPerformer
from mitosheet.step_performers.column_steps.delete_column import DeleteColumnStepPerformer
from mitosheet.step_performers.column_steps.rename_column import RenameColumnStepPerformer
from mitosheet.step_performers.column_steps.change_column_dtype import ChangeColumnDtypeStepPerformer
from mitosheet.step_performers.import_steps.simple_import import SimpleImportStepPerformer
from mitosheet.step_performers.dataframe_steps.dataframe_delete import DataframeDeleteStepPerformer
from mitosheet.step_performers.dataframe_steps.dataframe_duplicate import DataframeDuplicateStepPerformer
from mitosheet.step_performers.dataframe_steps.dataframe_rename import DataframeRenameStepPerformer
from mitosheet.step_performers.bulk_old_rename.bulk_old_rename import BulkOldRenameStepPerformer
from mitosheet.step_performers.graph_steps.graph import GraphStepPerformer
from mitosheet.step_performers.graph_steps.graph_delete import GraphDeleteStepPerformer
from mitosheet.step_performers.graph_steps.graph_duplicate import GraphDuplicateStepPerformer
from mitosheet.step_performers.graph_steps.graph_rename import GraphRenameStepPerformer
from mitosheet.step_performers.column_steps.split_text_to_columns import SplitTextToColumnsStepPerformer
from mitosheet.step_performers.fill_na import FillNaStepPerformer
from mitosheet.step_performers.delete_row import DeleteRowStepPerformer
from mitosheet.step_performers.promote_row_to_header import PromoteRowToHeaderStepPerformer
from mitosheet.step_performers.transpose import TransposeStepPerformer
from mitosheet.step_performers.melt import MeltStepPerformer
from mitosheet.enterprise.step_performers.one_hot_encoding import OneHotEncodingStepPerformer
from mitosheet.pro.step_performers.set_dataframe_format import SetDataframeFormatStepPerformer
from mitosheet.step_performers.import_steps.dataframe_import import DataframeImportStepPerformer
from mitosheet.step_performers.import_steps.snowflake_import import SnowflakeImportStepPerformer
from mitosheet.step_performers.import_steps.excel_range_import import ExcelRangeImportStepPerformer
from mitosheet.step_performers.export_to_file import ExportToFileStepPerformer
from mitosheet.step_performers.reset_index import ResetIndexStepPerformer
# AUTOGENERATED LINE: IMPORT (DO NOT DELETE)

# All steps must be listed in this variable. Note the Type annotation allows for
# subtypes of the step performer to be passed
STEP_PERFORMERS: List[Type[StepPerformer]] = [
    PivotStepPerformer,
    ReorderColumnStepPerformer,
    FilterStepPerformer,
    SortStepPerformer,
    SetCellValueStepPerformer,
    AddColumnStepPerformer,
    SetColumnFormulaStepPerformer,
    ChangeColumnDtypeStepPerformer,
    MergeStepPerformer,
    ConcatStepPerformer,
    FillNaStepPerformer,
    DeleteColumnStepPerformer,
    RenameColumnStepPerformer,
    SimpleImportStepPerformer,
    ExcelImportStepPerformer,
    DataframeDeleteStepPerformer,
    DataframeDuplicateStepPerformer,
    DataframeRenameStepPerformer,
    BulkOldRenameStepPerformer,
    DropDuplicatesStepPerformer,
    GraphStepPerformer,
    GraphDeleteStepPerformer,
    GraphDuplicateStepPerformer,
    GraphRenameStepPerformer,
    DeleteRowStepPerformer,
    PromoteRowToHeaderStepPerformer,
    SplitTextToColumnsStepPerformer,
    TransposeStepPerformer,
    MeltStepPerformer,
    OneHotEncodingStepPerformer,
    SetDataframeFormatStepPerformer,
    DataframeImportStepPerformer,
    SnowflakeImportStepPerformer,
    ExcelRangeImportStepPerformer,
    ExportToFileStepPerformer,
    ResetIndexStepPerformer,
    # AUTOGENERATED LINE: EXPORT (DO NOT DELETE)
]

# A helpful mapping for looking up steps based on the incoming events
EVENT_TYPE_TO_STEP_PERFORMER: Dict[str, Type[StepPerformer]] = {
    step_performer.step_event_type(): step_performer
    for step_performer in STEP_PERFORMERS
}

# We also build a useful lookup mapping for the step type to step object
STEP_TYPE_TO_STEP_PERFORMER: Dict[str, Type[StepPerformer]] = {
    step_performer.step_type(): step_performer
    for step_performer in STEP_PERFORMERS
}



#!/usr/bin/env python
# coding: utf-8

# Copyright (c) Saga Inc.
# Distributed under the terms of The Mito Enterprise license.

from time import perf_counter
from typing import Any, Dict, List, Optional, Set, Tuple

import pandas as pd
from mitosheet.code_chunks.code_chunk import CodeChunk
from mitosheet.code_chunks.one_hot_encoding_code_chunk import OneHotEncodingCodeChunk
from mitosheet.errors import make_column_exists_error, make_columns_exists_error

from mitosheet.state import State
from mitosheet.step_performers.step_performer import StepPerformer
from mitosheet.step_performers.utils import add_columns_to_df, get_param
from mitosheet.types import ColumnID

class OneHotEncodingStepPerformer(StepPerformer):
    """
    Allows you to do a one hot encoding of a column.
    """

    @classmethod
    def step_version(cls) -> int:
        return 1

    @classmethod
    def step_type(cls) -> str:
        return 'one_hot_encoding'

    @classmethod
    def execute(cls, prev_state: State, params: Dict[str, Any]) -> Tuple[State, Optional[Dict[str, Any]]]:
        sheet_index: int = get_param(params, 'sheet_index')
        column_id: ColumnID = get_param(params, 'column_id')

        # We make a new state to modify it
        post_state = prev_state.copy()

        column_header = prev_state.column_ids.get_column_header_by_id(sheet_index, column_id)
        df = post_state.dfs[sheet_index]
        column_index = df.columns.tolist().index(column_header)

        pandas_start_time = perf_counter()
        new_columns = pd.get_dummies(df[column_header])
        pandas_processing_time = perf_counter() - pandas_start_time

        # TODO: make sure there isn't overlap here?
        new_column_headers = new_columns.columns.tolist()

        # If there is overlap between the new column headers and the existing df, then
        # we throw an error about this overlap
        if len(set(new_column_headers) & set(df.columns.tolist())) > 0:
            raise make_columns_exists_error(set(new_column_headers) & set(df.columns.tolist()))
            
        final_df = add_columns_to_df(df, new_columns, new_column_headers, column_index)
        post_state.dfs[sheet_index] = final_df
        # Update column state variables
        post_state.add_columns_to_state(sheet_index, new_column_headers)

        return post_state, {
            'pandas_processing_time': pandas_processing_time,
            'new_column_headers': new_column_headers,
            'result': {
                # TODO: fill in the result
            }
        }

    @classmethod
    def transpile(
        cls,
        prev_state: State,
        post_state: State,
        params: Dict[str, Any],
        execution_data: Optional[Dict[str, Any]],
    ) -> List[CodeChunk]:
        return [
            OneHotEncodingCodeChunk(
                prev_state, 
                post_state, 
                get_param(params, 'sheet_index'), 
                get_param(params, 'column_id'), 
                get_param(execution_data if execution_data is not None else {}, 'new_column_headers')
            )
        ]

    @classmethod
    def get_modified_dataframe_indexes(cls, params: Dict[str, Any]) -> Set[int]:
        return {get_param(params, 'sheet_index')}
    



#!/usr/bin/env python
# coding: utf-8

# Copyright (c) Saga Inc.
# Distributed under the terms of the GPL License.
from copy import copy
from typing import Dict, List, Optional, Tuple

from mitosheet.code_chunks.code_chunk import CodeChunk
from mitosheet.excel_utils import (get_column_from_column_index,
                                   get_col_and_row_indexes_from_range)
from mitosheet.state import State


class ExcelRangeImportCodeChunk(CodeChunk):

    def __init__(self, prev_state: State, post_state: State, file_path: str, sheet_name: str, sheet_index_to_df_range: Dict[int, str]):
        super().__init__(prev_state, post_state)
        self.file_path = file_path
        self.sheet_name = sheet_name
        self.sheet_index_to_df_range = sheet_index_to_df_range

    def get_display_name(self) -> str:
        return 'Excel Range Import'
    
    def get_description_comment(self) -> str:
        
        return f"Imported {len(self.sheet_index_to_df_range)} dataframes from {self.sheet_name} in {self.file_path}"

    def get_code(self) -> Tuple[List[str], List[str]]:
        code = []
        for sheet_index, _range in self.sheet_index_to_df_range.items():
            df_name = self.post_state.df_names[sheet_index]
            ((start_col_index, start_row_index), (end_col_index, end_row_index)) = get_col_and_row_indexes_from_range(_range)
            nrows = end_row_index - start_row_index
            usecols = get_column_from_column_index(start_col_index) + ':' + get_column_from_column_index(end_col_index)
            
            code.append(
                f'{df_name} = pd.read_excel(\'{self.file_path}\', sheet_name=\'{self.sheet_name}\', skiprows={start_row_index}, nrows={nrows}, usecols=\'{usecols}\')'
            )

        return code, ['import pandas as pd']

    def get_created_sheet_indexes(self) -> Optional[List[int]]:
        return list(self.sheet_index_to_df_range.keys())

    def _combine_right_with_excel_range_import_code_chunk(self, excel_range_import_code_chunk: "ExcelRangeImportCodeChunk") -> Optional[CodeChunk]:
        if excel_range_import_code_chunk.file_path == self.file_path and excel_range_import_code_chunk.sheet_name == self.sheet_name:
            new_sheet_index_to_df_range = copy(self.sheet_index_to_df_range)
            new_sheet_index_to_df_range.update(excel_range_import_code_chunk.sheet_index_to_df_range)

            return ExcelRangeImportCodeChunk(
                self.prev_state,
                excel_range_import_code_chunk.post_state,
                self.file_path,
                self.sheet_name,
                new_sheet_index_to_df_range
            )

        return None


    def combine_right(self, other_code_chunk: "CodeChunk") -> Optional["CodeChunk"]:
        if isinstance(other_code_chunk, ExcelRangeImportCodeChunk):
            return self._combine_right_with_excel_range_import_code_chunk(other_code_chunk)

        return None
    


#!/usr/bin/env python
# coding: utf-8

# Copyright (c) Saga Inc.
# Distributed under the terms of the GPL License.
import json
from typing import Any, Dict

import pandas as pd
from mitosheet.code_chunks.step_performers.column_steps.split_text_to_columns_code_chunk import get_split_param_dict
from mitosheet.sheet_functions.types.utils import is_datetime_dtype, is_timedelta_dtype
from mitosheet.types import StepsManagerType
from mitosheet.utils import get_row_data_array

# The maximum number of values the front-end sends to the backend
# See comments in function description below.
MAX_UNIQUE_VALUES = 1_000

def get_split_text_to_columns_preview(params: Dict[str, Any], steps_manager: StepsManagerType) -> str:
    """
    Sends back a string that can be parsed to a JSON object that
    contains the first 3 rows of the new columns created by splittin the 
    column_id on the delimiters. 
    
    """
    sheet_index = params['sheet_index']
    column_id = params['column_id']
    delimiters = params['delimiters']

    column_header = steps_manager.curr_step.column_ids.get_column_header_by_id(sheet_index, column_id)
    
    # TODO: Make sure there are no NaN values in the first 3 rows, or at least one non-NaN value. 
    df_head = steps_manager.curr_step.dfs[sheet_index].head(3)
    delimiter_string = '|'.join(delimiters)

    split_param_dict = get_split_param_dict()

    # Create the dataframe of new columns. We do this first, so that we know how many columns get created.
    if is_datetime_dtype(str(df_head[column_header].dtype)):
        df_preview = df_head[column_header].dt.strftime('%Y-%m-%d %X').str.split(delimiter_string, **split_param_dict)
    elif is_timedelta_dtype(str(df_head[column_header].dtype)):
        df_preview = df_head[column_header].apply(lambda x: str(x)).str.split(delimiter_string, **split_param_dict)
    else:
        df_preview = df_head[column_header].astype('str').str.split(delimiter_string, **split_param_dict)

    return json.dumps({
        'dfPreviewRowDataArray': get_row_data_array(df_preview),
    })




#!/usr/bin/env python
# coding: utf-8

# Copyright (c) Saga Inc.
# Distributed under the terms of the GPL License.

import json
from typing import Any, Dict
import pandas as pd
from mitosheet.types import StepsManagerType


NO_DEFINED_DF_MESSAGE = 'No variables match your requested type.'

def get_df_names() -> Dict[str, pd.DataFrame]:
    from IPython import get_ipython
    from io import StringIO 
    import sys

    ipython = get_ipython() # type: ignore

    class Capturing(list):
        def __enter__(self):
            self._stdout = sys.stdout
            sys.stdout = self._stringio = StringIO()
            return self
        def __exit__(self, *args):
            self.extend(self._stringio.getvalue().splitlines())
            del self._stringio    # free up some memory
            sys.stdout = self._stdout

    with Capturing() as output:
        ipython.run_line_magic("who",  "DataFrame")
        
    output = [df.strip() for line in output for df in line.split("\t") if df.strip() != '']
    
    # If we get a message there are no variables, clear this out of the ouput
    if len(output) == 1 and output[0] == NO_DEFINED_DF_MESSAGE:
        output = []

    return output


def get_defined_df_names(params: Dict[str, Any], steps_manager: StepsManagerType) -> str:
    return json.dumps(get_df_names())



#!/usr/bin/env python
# coding: utf-8

# Copyright (c) Saga Inc.
# Distributed under the terms of the GPL License.
from collections import OrderedDict
from copy import deepcopy
from typing import Any, Collection, List, Dict, Optional
import pandas as pd

from mitosheet.column_headers import ColumnIDMap
from mitosheet.types import FrontendFormulaAndLocation
from mitosheet.types import ColumnHeader, ColumnID, DataframeFormat
from mitosheet.utils import get_first_unused_dataframe_name

# Constants for where the dataframe in the state came from
DATAFRAME_SOURCE_PASSED = "passed"  # passed in mitosheet.sheet
DATAFRAME_SOURCE_IMPORTED = "imported"  # imported through a simple import
DATAFRAME_SOURCE_PIVOTED = "pivoted"  # created through a pivot
DATAFRAME_SOURCE_MERGED = "merged"  # created through a merge
DATAFRAME_SOURCE_CONCAT = "concated"  # created through a concat
DATAFRAME_SOURCE_DUPLICATED = "duplicated"  # created through a sheet duplication
DATAFRAME_SOURCE_TRANSPOSED = "transposed"  # created through a dataframe being transposed
DATAFRAME_SOURCE_MELTED = "melted"  # created through a dataframe being melted

# Constants used for formatting. Defined here to avoid circular imports
NUMBER_FORMAT_PLAIN_TEXT = "plain text"
NUMBER_FORMAT_CURRENCY = "currency"
NUMBER_FORMAT_ACCOUNTING = "accounting"
NUMBER_FORMAT_PERCENTAGE = "percentage"
NUMBER_FORMAT_SCIENTIFIC_NOTATION = "scientific notation"


def get_default_dataframe_format() -> DataframeFormat:
    return {
        "columns": {},
        "headers": {},
        "rows": {
            "even": {},
            "odd": {},
        },
        "border": {},
        "conditional_formats": []
    }


class State:
    """
    State is a container that stores the current state of a Mito analysis,
    where each step that is applied takes a state as input and creates a
    new state as output.

    It stores the obvious things, like the dataframes and their names, but
    also other helper pieces of state like: the column formulas, the filters,
    etc.
    """

    def __init__(
        self,
        dfs: Collection[pd.DataFrame],
        df_names: Optional[List[str]]=None,
        df_sources: Optional[List[str]]=None,
        column_ids: Optional[ColumnIDMap]=None,
        column_formulas: Optional[List[Dict[ColumnID, List[FrontendFormulaAndLocation]]]]=None,
        column_filters: Optional[List[Dict[ColumnID, Any]]]=None,
        df_formats: Optional[List[DataframeFormat]]=None,
        graph_data_dict: "Optional[OrderedDict[str, Dict[str, Any]]]"=None
    ):

        # The dataframes that are in the state
        self.dfs = list(dfs)

        # The df_names are composed of two parts:
        # 1. The names of the variables passed into the mitosheet.sheet call (which don't change over time).
        # 2. The names of the dataframes that were created during the analysis (e.g. by a merge).
        # Until we get them from the frontend as an update_event, we default them to df1, df2, ...
        self.df_names = (
            df_names
            if df_names is not None
            else [f"df{i + 1}" for i in range(len(dfs))]
        )

        # The df sources are where the actual dataframes come from, e.g.
        # how the dataframe was created. If not df sources passed, then this is in the
        # initialize state, and so these dataframes were passed to the mitosheet
        # call
        self.df_sources = (
            df_sources
            if df_sources is not None
            else [DATAFRAME_SOURCE_PASSED for _ in dfs]
        )

        # We then make a column id map if we do not already have one, so that we can identify each
        # of the columns from their static ids through the rest of the analysis.
        # NOTE: every state variable below that is defined per column is access by _ids_ not
        # by column headers. The column headers _only_ index into the dataframe itself
        self.column_ids = column_ids if column_ids else ColumnIDMap(dfs)

        self.column_formulas: List[Dict[ColumnID, List[FrontendFormulaAndLocation]]] = (
            column_formulas
            if column_formulas is not None
            else [
                {
                    column_id: []
                    for column_id in self.column_ids.get_column_ids(sheet_index)
                }
                for sheet_index in range(len(dfs))
            ]
        )

        self.column_filters = (
            column_filters
            if column_filters is not None
            else [
                {
                    column_id: {"operator": "And", "filters": []}
                    for column_id in self.column_ids.get_column_ids(sheet_index)
                }
                for sheet_index in range(len(dfs))
            ]
        )

        self.df_formats: List[DataframeFormat] = (
            df_formats
            if df_formats is not None
            else [
                get_default_dataframe_format()
                for _ in range(len(dfs))
            ]
        )

        # We put this in an ordered dict so we can easily figure out the last graph that was edited at each step. 
        # This is helpful for undoing, for example. 
        self.graph_data_dict: OrderedDict[str, Dict[str, Any]] = graph_data_dict if graph_data_dict is not None else OrderedDict()

    def copy(self, deep_sheet_indexes: Optional[List[int]]=None) -> "State":
        """
        Returns a copy of the state, while only making deep copies of
        those dataframes in the deep_sheet_indexes. Ideally, we'd copy
        even less than that deeply.
        """
        if deep_sheet_indexes is None:
            deep_sheet_indexes = []
        
        return State(
            [df.copy(deep=index in deep_sheet_indexes) for index, df in enumerate(self.dfs)],
            df_names=deepcopy(self.df_names),
            df_sources=deepcopy(self.df_sources),
            column_ids=deepcopy(self.column_ids),
            column_formulas=deepcopy(self.column_formulas),
            column_filters=deepcopy(self.column_filters),
            df_formats=deepcopy(self.df_formats),
            graph_data_dict=deepcopy(self.graph_data_dict)
        )

    def add_df_to_state(
        self,
        new_df: pd.DataFrame,
        df_source: str,
        sheet_index: Optional[int]=None,
        df_name: Optional[str]=None,
        df_format: Optional[DataframeFormat]=None,
        use_deprecated_id_algorithm: bool=False, 
    ) -> int:
        """
        Helper function for adding a new dataframe to this state,
        and keeping all the other variables in sync.

        If sheet_index is defined, then will replace the dataframe
        that is currently at the index. Otherwise, if sheet_index is
        not defined, then will append the df to the end of the state
        """
        if sheet_index is None:
            # Update dfs by appending new df
            self.dfs.append(new_df)
            # Also update the dataframe name
            if df_name is None:
                self.df_names.append(
                    get_first_unused_dataframe_name(self.df_names, f"df{len(self.df_names) + 1}")
                )
            else:
                self.df_names.append(df_name)

            # Save the source of this dataframe
            self.df_sources.append(df_source)

            # Add this to the column_ids map
            column_ids = self.column_ids.add_df(
                new_df, use_deprecated_id_algorithm=use_deprecated_id_algorithm
            )

            # Update all the variables that depend on column_headers
            self.column_formulas.append(
                {column_id: [] for column_id in column_ids}
            )
            self.column_filters.append(
                {
                    column_id: {"operator": "And", "filters": []}
                    for column_id in column_ids
                }
            )
            self.df_formats.append(
                get_default_dataframe_format()
                if df_format is None
                else df_format
            )

            # Return the index of this sheet
            return len(self.dfs) - 1
        else:

            # Update dfs by switching which df is at this index specifically
            self.dfs[sheet_index] = new_df
            # Also update the dataframe name, if it is passed. Otherwise, we don't change it
            if df_name is not None:
                self.df_names[sheet_index] = df_name

            # Save the source of this dataframe, if it is passed. Otherwise, don't change it
            if df_source is not None:
                self.df_sources[sheet_index] = df_source

            # Add this to the column_ids map
            column_ids = self.column_ids.add_df(
                new_df,
                sheet_index=sheet_index,
                use_deprecated_id_algorithm=use_deprecated_id_algorithm,
            )

            # Update all the variables that depend on column_headers
            self.column_formulas[sheet_index] = {
                column_id: [] for column_id in column_ids
            }
            self.column_filters[sheet_index] = {
                column_id: {"operator": "And", "filters": []}
                for column_id in column_ids
            }
            self.df_formats[sheet_index] = (
                get_default_dataframe_format()
                if df_format is None
                else df_format
            )

            # Return the index of this sheet
            return sheet_index

    def add_columns_to_state(self, sheet_index: int, column_headers: List[ColumnHeader]) -> None:
        """
        Helper function for adding a new columns to this state, making sure that we 
        track the relevant metadata variables.
        """
        # Update column state variables
        for column_header in column_headers:
            column_id = self.column_ids.add_column_header(sheet_index, column_header)
            self.column_formulas[sheet_index][column_id] = []
            self.column_filters[sheet_index][column_id] = {'operator': 'And', 'filters': []}

    def does_sheet_index_exist_within_state(self, sheet_index: int) -> bool:
        """
        Returns true iff a sheet_index exists within this state
        """
        return not (sheet_index < 0 or sheet_index >= len(self.dfs))

    def move_to_deprecated_id_algorithm(self) -> None:
        """
        This helper function will move the entire state to the new IDs,
        for backwards compatibility reasons. Namely, users who pass
        dataframes to mito directly have a bulk_old_rename appended to
        the start of their analysis that calls this function.

        This moves the IDs to what they would have been if there was a
        preprocessing step that had run, that had performed renames on
        the column headers before they were allowed into Mito. Brutal.
        """
        from mitosheet.step_performers.bulk_old_rename.deprecated_utils import (
            make_valid_header,
        )

        # Loop over all the attributes of this object
        for key, value in self.__dict__.items():
            # And for anything defined on columns, update it to the new id schema
            if key.startswith("column") and key != "column_ids":
                new_value = [
                    {make_valid_header(k): v for k, v in column_map.items()}
                    for column_map in value
                ]
                self.__setattr__(key, new_value)

        # Then, update the column ids mapping object itself
        self.column_ids.move_to_deprecated_id_format()



import pandas as pd
from mitosheet.sheet_functions.types.utils import is_int_dtype, is_number_dtype
from mitosheet.state import (NUMBER_FORMAT_ACCOUNTING,
                             NUMBER_FORMAT_PERCENTAGE, NUMBER_FORMAT_PLAIN_TEXT, NUMBER_FORMAT_SCIENTIFIC_NOTATION)
from mitosheet.types import StepsManagerType

from mitosheet.excel_utils import get_excel_range_from_column_index


def add_formatting_to_excel_sheet(
        writer: pd.ExcelWriter, 
        steps_manager: StepsManagerType, 
        sheet_index: int,
    ) -> None:
    """
    Adds formatting to the sheet_name, based on the formatting the user
    currently has applied in the frontend. 

    NOTE: this is a Mito Pro feature.
    """
    df = steps_manager.dfs[sheet_index]
    sheet_name = steps_manager.curr_step.df_names[sheet_index]

    # TODO: fix up below
    return

    format_data_map = steps_manager.curr_step.column_format_types[sheet_index]

    workbook  = writer.book
    worksheet = writer.sheets[sheet_name]

    ## This formatting code is wrapped in a if False until we deploy our first pro features
    for column_index, column_header in enumerate(df.keys()):
        column_id = steps_manager.curr_step.column_ids.get_column_id_by_header(sheet_index, column_header)
        column_dtype = str(df[column_header].dtype)
        format_data = format_data_map[column_id]
        excel_column = get_excel_range_from_column_index(column_index)

        if (is_number_dtype(column_dtype)):
            format = workbook.add_format({'num_format': 'General'})
            worksheet.set_column(excel_column, None, format)


#!/usr/bin/env python
# coding: utf-8

# Copyright (c) Saga Inc.
# Distributed under the terms of the GPL License.
from mitosheet.transpiler.transpile_utils import NEWLINE_TAB
import pytest
import pandas as pd

from mitosheet.transpiler.transpile import transpile
from mitosheet.tests.test_utils import create_mito_wrapper, create_mito_wrapper_dfs


def test_transpile_single_column():
    mito = create_mito_wrapper(['abc'])
    mito.set_formula('=A', 0, 'B', add_column=True)

    assert mito.transpiled_code == [
        "df1.insert(1, 'B', df1[\'A\'])", 
    ]


def test_transpile_multiple_columns_no_relationship():
    mito = create_mito_wrapper(['abc'])
    mito.add_column(0, 'B')
    mito.add_column(0, 'C')
    assert mito.transpiled_code == [
        'df1.insert(1, \'B\', 0)',
        'df1.insert(2, \'C\', 0)',
    ]

def test_transpile_columns_in_each_sheet():
    mito = create_mito_wrapper(['abc'], sheet_two_A_data=['abc'])
    mito.add_column(0, 'B')
    mito.add_column(1, 'B')

    assert mito.transpiled_code == [
        'df1.insert(1, \'B\', 0)',
        'df2.insert(1, \'B\', 0)',
    ]

def test_transpile_multiple_columns_linear():
    mito = create_mito_wrapper(['abc'])
    mito.set_formula('=A', 0, 'B', add_column=True)
    mito.set_formula('=B', 0, 'C', add_column=True)

    assert mito.transpiled_code == [
        'df1.insert(1, \'B\', df1[\'A\'])',
        'df1.insert(2, \'C\', df1[\'B\'])',
    ]

COLUMN_HEADERS = [
    ('ABC'),
    ('ABC_D'),
    ('ABC_DEF'),
    ('ABC_123'),
    ('ABC_HAHA_123'),
    ('ABC_HAHA-123'),
    ('---data---'),
    ('---da____ta---'),
    ('--'),
]
@pytest.mark.parametrize("column_header", COLUMN_HEADERS)
def test_transpile_column_headers_non_alphabet(column_header):
    mito = create_mito_wrapper(['abc'])
    mito.set_formula('=A', 0, column_header, add_column=True)

    assert mito.transpiled_code == [
        f'df1.insert(1, \'{column_header}\', df1[\'A\'])', 
    ]


COLUMN_HEADERS = [
    ('ABC'),
    ('ABC_D'),
    ('ABC_DEF'),
    ('ABC_123'),
    ('ABC_HAHA_123'),
    ('ABC_HAHA-123'),
    ('---data---'),
    ('---da____ta---'),
    ('--'),
]
@pytest.mark.parametrize("column_header", COLUMN_HEADERS)
def test_transpile_column_headers_non_alphabet_multi_sheet(column_header):
    mito = create_mito_wrapper(['abc'], sheet_two_A_data=['abc'])
    mito.set_formula('=A', 0, column_header, add_column=True)
    mito.set_formula('=A', 1, column_header, add_column=True)

    assert mito.transpiled_code == [
        f'df1.insert(1, \'{column_header}\', df1[\'A\'])', 
        f'df2.insert(1, \'{column_header}\', df2[\'A\'])', 
    ]

def test_preserves_order_columns():
    mito = create_mito_wrapper(['abc'])
    # Topological sort will currently display this in C, B order
    mito.add_column(0, 'B')
    mito.add_column(0, 'C')

    assert mito.transpiled_code == [
        'df1.insert(1, \'B\', 0)',
        'df1.insert(2, \'C\', 0)',
    ]

def test_transpile_delete_columns():
    df1 = pd.DataFrame(data={'A': [1], 'B': [101], 'C': [11]})
    mito = create_mito_wrapper_dfs(df1)
    mito.delete_columns(0, ['C', 'B'])

    assert mito.transpiled_code == [
        'df1.drop([\'C\', \'B\'], axis=1, inplace=True)'
    ]


# TESTING OPTIMIZATION

def test_removes_unedited_formulas_for_unedited_sheets():
    df1 = pd.DataFrame(data={'A': [1], 'B': [101], 'C': [11]})
    df2 = pd.DataFrame(data={'A': [1], 'B': [101], 'C': [11]})
    mito = create_mito_wrapper_dfs(df1, df2)

    mito.set_formula('=C', 0, 'D', add_column=True)
    mito.set_formula('=C', 1, 'D', add_column=True)

    mito.merge_sheets('lookup', 0, 1, [['A', 'A']], ['A', 'B', 'C', 'D'], ['A', 'B', 'C', 'D'])

    mito.set_formula('=C + 1', 1, 'D', add_column=True)

    assert mito.transpiled_code == [
        "df1.insert(3, 'D', df1[\'C\'])", 
        "df2.insert(3, 'D', df2[\'C\'])", 
        'temp_df = df2.drop_duplicates(subset=[\'A\']) # Remove duplicates so lookup merge only returns first match', 
        'df3 = df1.merge(temp_df, left_on=[\'A\'], right_on=[\'A\'], how=\'left\', suffixes=[\'_df1\', \'_df2\'])',
        'df2[\'D\'] = df2[\'C\'] + 1',
    ]


def test_mulitple_merges_no_formula_steps():
    df1 = pd.DataFrame(data={'A': [1], 'B': [101], 'C': [11]})
    df2 = pd.DataFrame(data={'A': [1], 'B': [101], 'C': [11]})
    mito = create_mito_wrapper_dfs(df1, df2)
    mito.merge_sheets('lookup', 0, 1, [['A', 'A']], ['A', 'B', 'C'], ['A', 'B', 'C'])
    mito.merge_sheets('lookup', 0, 1, [['A', 'A']], ['A', 'B', 'C'], ['A', 'B', 'C'])
    mito.merge_sheets('lookup', 0, 1, [['A', 'A']], ['A', 'B', 'C'], ['A', 'B', 'C'])


    assert mito.transpiled_code == [
        'temp_df = df2.drop_duplicates(subset=[\'A\']) # Remove duplicates so lookup merge only returns first match', 
        'df3 = df1.merge(temp_df, left_on=[\'A\'], right_on=[\'A\'], how=\'left\', suffixes=[\'_df1\', \'_df2\'])',
        'temp_df = df2.drop_duplicates(subset=[\'A\']) # Remove duplicates so lookup merge only returns first match', 
        'df4 = df1.merge(temp_df, left_on=[\'A\'], right_on=[\'A\'], how=\'left\', suffixes=[\'_df1\', \'_df2\'])',
        'temp_df = df2.drop_duplicates(subset=[\'A\']) # Remove duplicates so lookup merge only returns first match', 
        'df5 = df1.merge(temp_df, left_on=[\'A\'], right_on=[\'A\'], how=\'left\', suffixes=[\'_df1\', \'_df2\'])'
    ]

def test_optimization_with_other_edits():
    df1 = pd.DataFrame(data={'A': [1], 'B': [101], 'C': [11]})
    df2 = pd.DataFrame(data={'A': [1], 'B': [101], 'C': [11]})
    mito = create_mito_wrapper_dfs(df1, df2)
    mito.add_column(0, 'D')
    mito.set_formula('=A', 0, 'D')
    mito.merge_sheets('lookup', 0, 1, [['A', 'A']], ['A', 'B', 'C', 'D'], ['A', 'B', 'C'])
    mito.add_column(0, 'AAA')
    mito.delete_columns(0, ['AAA'])

    assert mito.transpiled_code == [
        "df1.insert(3, 'D', df1[\'A\'])", 
        'temp_df = df2.drop_duplicates(subset=[\'A\']) # Remove duplicates so lookup merge only returns first match', 
        'df3 = df1.merge(temp_df, left_on=[\'A\'], right_on=[\'A\'], how=\'left\', suffixes=[\'_df1\', \'_df2\'])',
    ]

flatten_code = f'pivot_table.columns = [{NEWLINE_TAB}\'_\'.join([str(c) for c in col]).strip() if isinstance(col, tuple) else col{NEWLINE_TAB}for col in pivot_table.columns.values\n]'


def test_transpile_does_no_initial():
    df1 = pd.DataFrame(data={'First Name': ['Nate', 'Nate'], 123: ['Rush', 'Jack'], True: ['1', '2']})
    mito = create_mito_wrapper_dfs(df1)

    assert len(mito.transpiled_code) == 0

    
def test_transpile_reorder_column():
    df1 = pd.DataFrame(data={'A': ['aaron'], 'B': ['jon']})
    mito = create_mito_wrapper_dfs(df1)
    mito.reorder_column(0, 'A', 1)

    assert mito.transpiled_code == [
        'df1_columns = [col for col in df1.columns if col != \'A\']',
        'df1_columns.insert(1, \'A\')',
        'df1 = df1[df1_columns]'
    ]

def test_transpile_two_column_reorders():
    df1 = pd.DataFrame(data={'A': ['aaron'], 'B': ['jon']})
    mito = create_mito_wrapper_dfs(df1)
    mito.reorder_column(0, 'A', 1)
    mito.reorder_column(0, 'B', 1)

    assert mito.transpiled_code == [
        'df1_columns = [col for col in df1.columns if col != \'A\']',
        'df1_columns.insert(1, \'A\')',
        'df1 = df1[df1_columns]',
        'df1_columns = [col for col in df1.columns if col != \'B\']',
        'df1_columns.insert(1, \'B\')',
        'df1 = df1[df1_columns]'
    ]

def test_transpile_reorder_column_invalid():
    df1 = pd.DataFrame(data={'A': ['aaron'], 'B': ['jon']})
    mito = create_mito_wrapper_dfs(df1)
    mito.reorder_column(0, 'A', 5)

    assert mito.transpiled_code == [
        'df1_columns = [col for col in df1.columns if col != \'A\']',
        'df1_columns.insert(1, \'A\')',
        'df1 = df1[df1_columns]',
    ]

def test_transpile_merge_then_sort():
    df1 = pd.DataFrame(data={'Name': ["Aaron", "Nate"], 'Number': [123, 1]})
    df2 = pd.DataFrame(data={'Name': ["Aaron", "Nate"], 'Sign': ['Gemini', "Tarus"]})
    mito = create_mito_wrapper_dfs(df1, df2)
    mito.merge_sheets('lookup', 0, 1, [['Name', 'Name']], list(df1.keys()), list(df2.keys()))
    mito.sort(2, 'Number', 'ascending')

    assert mito.transpiled_code == [
        'temp_df = df2.drop_duplicates(subset=[\'Name\']) # Remove duplicates so lookup merge only returns first match',
        'df3 = df1.merge(temp_df, left_on=[\'Name\'], right_on=[\'Name\'], how=\'left\', suffixes=[\'_df1\', \'_df2\'])',
        'df3 = df3.sort_values(by=\'Number\', ascending=True, na_position=\'first\')',
    ]

def test_transpile_multiple_pandas_imports_combined(tmp_path):
    tmp_file = str(tmp_path / 'txt.csv')
    df1 = pd.DataFrame(data={'Name': ["Aaron", "Nate"], 'Number': [123, 1]})
    df1.to_csv(tmp_file, index=False)
    mito = create_mito_wrapper_dfs(df1)
    mito.simple_import([tmp_file])
    mito.add_column(0, 'A', -1)
    mito.simple_import([tmp_file])
    mito.add_column(1, 'A', -1)
    mito.simple_import([tmp_file])

    assert len(mito.optimized_code_chunks) == 5
    assert 'import pandas as pd' in mito.transpiled_code
    assert len([c for c in mito.transpiled_code if c == 'import pandas as pd']) == 1





#!/usr/bin/env python
# coding: utf-8

# Copyright (c) Saga Inc.
# Distributed under the terms of the GPL License.

from time import perf_counter
from typing import Any, Dict, List, Optional, Set, Tuple
from mitosheet.code_chunks.code_chunk import CodeChunk
from mitosheet.code_chunks.export_to_file_code_chunk import ExportToFileCodeChunk
from mitosheet.excel_utils import get_df_name_as_valid_sheet_name
import pandas as pd
import os

from mitosheet.state import State
from mitosheet.step_performers.step_performer import StepPerformer
from mitosheet.step_performers.utils import get_param

def get_export_to_csv_sheet_index_to_file_name(state: State, file_name: str, sheet_indexes: List[int]) -> Dict[int, str]:
    if len(sheet_indexes) == 1:
        return {sheet_indexes[0]: f'{file_name}'}
    else:
        file_name_without_extension, extension = os.path.splitext(file_name)
        return {sheet_index: f'{file_name_without_extension}_{state.df_names[sheet_index]}{extension}' for sheet_index in sheet_indexes}

def get_export_to_excel_sheet_index_to_sheet_name(state: State, file_name: str, sheet_indexes: List[int]) -> Dict[int, str]:
    return {sheet_index: get_df_name_as_valid_sheet_name(state.df_names[sheet_index]) for sheet_index in sheet_indexes}


def get_final_file_name(file_name: str, _type: str) -> str:
    _, extension = os.path.splitext(file_name)
    if extension == '':
        if _type == 'csv':
            return file_name + '.csv'
        elif _type == 'excel':
            return file_name + '.xlsx'
    
    return file_name


class ExportToFileStepPerformer(StepPerformer):
    """
    Allows you to export to file.
    """

    @classmethod
    def step_version(cls) -> int:
        return 1

    @classmethod
    def step_type(cls) -> str:
        return 'export_to_file'

    @classmethod
    def execute(cls, prev_state: State, params: Dict[str, Any]) -> Tuple[State, Optional[Dict[str, Any]]]:
        _type: str = get_param(params, 'type')
        sheet_indexes: List[int] = get_param(params, 'sheet_indexes')
        _file_name: str = get_param(params, 'file_name')

        file_name = get_final_file_name(_file_name, _type) # Ensure that the file name has the correct extension
        
        # We make a new state to modify it
        post_state = prev_state.copy()

        pandas_start_time = perf_counter()

        if _type == 'csv':
            sheet_index_to_export_location = get_export_to_csv_sheet_index_to_file_name(post_state, file_name, sheet_indexes)
            for sheet_index, file_name in sheet_index_to_export_location.items():
                post_state.dfs[sheet_index].to_csv(file_name, index=False)
        elif _type == 'excel':
            sheet_index_to_export_location = get_export_to_excel_sheet_index_to_sheet_name(post_state, file_name, sheet_indexes)
            with pd.ExcelWriter(file_name) as writer:
                for sheet_index, sheet_name in sheet_index_to_export_location.items():
                    post_state.dfs[sheet_index].to_excel(writer, sheet_name=sheet_name, index=False)
        else:
            raise ValueError(f"Invalid file type: {_type}")

        pandas_processing_time = perf_counter() - pandas_start_time

        return post_state, {
            'pandas_processing_time': pandas_processing_time,
            'sheet_index_to_export_location': sheet_index_to_export_location,
            'file_name': file_name,
            'result': {}
        }

    @classmethod
    def transpile(
        cls,
        prev_state: State,
        post_state: State,
        params: Dict[str, Any],
        execution_data: Optional[Dict[str, Any]],
    ) -> List[CodeChunk]:
        return [
            ExportToFileCodeChunk(
                prev_state, 
                post_state, 
                get_param(params, 'type'),
                get_param(execution_data if execution_data is not None else {}, 'file_name'),
                get_param(execution_data if execution_data is not None else {}, 'sheet_index_to_export_location'),
            )
        ]

    @classmethod
    def get_modified_dataframe_indexes(cls, params: Dict[str, Any]) -> Set[int]:
        return set()
    


#!/usr/bin/env python
# coding: utf-8

# Copyright (c) Saga Inc.
# Distributed under the terms of the GPL License.

from copy import copy
from typing import Any, Dict, List, Optional, Tuple

from mitosheet.code_chunks.code_chunk import CodeChunk
from mitosheet.code_chunks.step_performers.column_steps.delete_column_code_chunk import DeleteColumnsCodeChunk
from mitosheet.state import State
from mitosheet.transpiler.transpile_utils import column_header_map_to_string
from mitosheet.types import ColumnHeader, ColumnID


class RenameColumnsCodeChunk(CodeChunk):

    def __init__(self, prev_state: State, post_state: State, sheet_index: int, column_ids_to_new_column_headers: Dict[ColumnID, str]):
        super().__init__(prev_state, post_state)
        self.sheet_index = sheet_index
        self.column_ids_to_new_column_headers = column_ids_to_new_column_headers

    def get_display_name(self) -> str:
        return 'Renamed columns'
    
    def get_description_comment(self) -> str:
        new_column_headers = self.column_ids_to_new_column_headers.values()
        return f'Renamed columns {", ".join(new_column_headers)}'

    def get_code(self) -> Tuple[List[str], List[str]]:

        old_column_header_to_new_column_header_map: Dict[ColumnHeader, ColumnHeader] = dict()
        df_name = self.post_state.df_names[self.sheet_index]
        for column_id, new_column_header in self.column_ids_to_new_column_headers.items():
            old_column_header_to_new_column_header_map[self.prev_state.column_ids.get_column_header_by_id(self.sheet_index, column_id)] = new_column_header

        rename_dict = column_header_map_to_string(old_column_header_to_new_column_header_map)
        rename_string = f'{df_name}.rename(columns={rename_dict}, inplace=True)'
        return [rename_string], []

    def get_edited_sheet_indexes(self) -> List[int]:
        return [self.sheet_index]

    def _combine_right_with_rename_columns_code_chunk(self, other_code_chunk: "RenameColumnsCodeChunk") -> Optional["RenameColumnsCodeChunk"]:
        if not self.params_match(other_code_chunk, ['sheet_index']):
            return None

        new_rename_dict = copy(self.column_ids_to_new_column_headers)
        new_rename_dict.update(
            other_code_chunk.column_ids_to_new_column_headers
        )

        return RenameColumnsCodeChunk(
            self.prev_state,
            other_code_chunk.post_state,
            self.sheet_index,
            new_rename_dict
        )
    
    def _combine_right_with_delete_columns_code_chunk(self, other_code_chunk: "DeleteColumnsCodeChunk") -> Optional[CodeChunk]:
        if not self.params_match(other_code_chunk, ['sheet_index']):
            return None

        column_ids_to_new_column_headers = self.column_ids_to_new_column_headers
        column_ids_being_renamed = column_ids_to_new_column_headers.keys()
        column_ids_being_deleted = other_code_chunk.column_ids

        if set(column_ids_being_renamed).issubset(set(column_ids_being_deleted)):
            # If the columns being deleted are columns that were renamed, we can skip
            # the rename step and just do the deletion of these columns
            return DeleteColumnsCodeChunk(
                self.prev_state,
                other_code_chunk.post_state,
                other_code_chunk.sheet_index,
                other_code_chunk.column_ids,
            )
        
        return None

    def combine_right(self, other_code_chunk: CodeChunk) -> Optional[CodeChunk]:
        if isinstance(other_code_chunk, RenameColumnsCodeChunk):
            return self._combine_right_with_rename_columns_code_chunk(other_code_chunk)
        if isinstance(other_code_chunk, DeleteColumnsCodeChunk):
            return self._combine_right_with_delete_columns_code_chunk(other_code_chunk)
            
        return None


#!/usr/bin/env python
# coding: utf-8

# Copyright (c) Saga Inc.
# Distributed under the terms of the GPL License.
from copy import copy
from typing import Any, Dict, List
from mitosheet.step import Step
from mitosheet.types import StepsManagerType
from mitosheet.utils import create_step_id


UPDATE_EXISTING_IMPORTS_UPDATE_EVENT = 'update_existing_import_update'
UPDATE_EXISTING_IMPORTS_PARAMS = ['updated_step_import_data_list']


def execute_update_existing_imports_update(steps_manager: StepsManagerType, updated_step_import_data_list: List[Dict[str, Any]]) -> None:
    """
    Updates the steps by overwriting all existing imports with with 
    updated_step_import_data_list. 

    Note that the params in updated_step_import_data_list should each only 
    contain a single dataframe creation, as this is what is created by the
    frontend for convenience.
    """
    old_steps = copy(steps_manager.steps_including_skipped)
    new_steps = copy(steps_manager.steps_including_skipped)

    # We group up all the steps by their ids
    imports_for_step_id: Dict[str, List[Dict[str, Any]]] = dict()
    for step_import_data in updated_step_import_data_list:
        step_id = step_import_data['step_id']
        imports = step_import_data['imports']

        imports_for_step_id[step_id] = imports_for_step_id.get(step_id, []) + imports

    for step_id, imports in imports_for_step_id.items():
        original_step_index = [index for index, step in enumerate(new_steps) if step.step_id == step_id][0]

        # Build all the new steps, and append them to the new step list in place of the old import
        import_steps_to_replace_with = []
        for _import in imports:
            import_steps_to_replace_with.append(
                Step(_import['step_type'], create_step_id(), _import["params"])
            )

        # Then, replace the single old step with all the new steps
        new_steps = new_steps[:original_step_index] + import_steps_to_replace_with + new_steps[(original_step_index + 1):]

    try:
        # Refresh the anlaysis starting from the first step
        steps_manager.execute_and_update_steps(new_steps, 0)
    except Exception as e:
        # We also make sure to catch any error, and turn it into an error that
        # does not go directly to an error modal.
        from mitosheet.errors import MitoError, make_invalid_update_imports_error
        if isinstance(e, MitoError):
            e.error_modal = False
            raise e
        
        raise make_invalid_update_imports_error()

    steps_manager.undone_step_list_store.append(("reset", old_steps))


UPDATE_EXISTING_IMPORTS_UPDATE = {
    'event_type': UPDATE_EXISTING_IMPORTS_UPDATE_EVENT,
    'params': UPDATE_EXISTING_IMPORTS_PARAMS,
    'execute': execute_update_existing_imports_update
}


#!/usr/bin/env python
# coding: utf-8

# Copyright (c) Saga Inc.
# Distributed under the terms of the GPL License.
from typing import Any, Dict, List, Optional, Tuple

from mitosheet.code_chunks.code_chunk import CodeChunk
from mitosheet.code_chunks.code_chunk_utils import get_right_combine_with_column_delete_code_chunk
from mitosheet.code_chunks.step_performers.column_steps.delete_column_code_chunk import DeleteColumnsCodeChunk
from mitosheet.sheet_functions.types.utils import (is_bool_dtype,
                                                   is_datetime_dtype,
                                                   is_int_dtype,
                                                   is_number_dtype,
                                                   is_timedelta_dtype)
from mitosheet.state import State
from mitosheet.transpiler.transpile_utils import \
    column_header_to_transpiled_code
from mitosheet.types import ColumnID


class SetCellValueCodeChunk(CodeChunk):

    def __init__(self, prev_state: State, post_state: State, sheet_index: int, column_id: ColumnID, row_index: Any, old_value: Any, new_value: Any, type_corrected_new_value: Optional[Any]):
        super().__init__(prev_state, post_state)
        self.sheet_index = sheet_index
        self.column_id = column_id
        self.row_index = row_index
        self.old_value = old_value
        self.new_value = new_value
        self.type_corrected_new_value = type_corrected_new_value

        self.df_name = self.post_state.df_names[self.sheet_index]

    def get_display_name(self) -> str:
        return 'Set cell value'
    
    def get_description_comment(self) -> str:
        column_header = self.post_state.column_ids.get_column_header_by_id(self.sheet_index, self.column_id)
        return f'Set a cell value in {column_header}'

    def get_code(self) -> Tuple[List[str], List[str]]:

        code: List[str] = []

        # If nothings changed, we don't write any code
        if self.old_value == self.new_value:
            return code, []

        column_header = self.post_state.column_ids.get_column_header_by_id(self.sheet_index, self.column_id)
        transpiled_column_header = column_header_to_transpiled_code(column_header)

        # If the series is an int, but the new value is a float, convert the series to floats before adding the new value
        column_dtype = str(self.prev_state.dfs[self.sheet_index][column_header].dtype)
        if self.new_value is not None and '.' in self.new_value and is_int_dtype(column_dtype):
            code.append(f'{self.df_name}[{transpiled_column_header}] = {self.df_name}[\'{column_header}\'].astype(\'float\')')

        # Actually set the new value
        # We don't need to wrap the value in " if its None, a Boolean Series, or a Number Series.
        if self.type_corrected_new_value is None or is_bool_dtype(column_dtype) or is_number_dtype(column_dtype):
            code.append(f'{self.df_name}.at[{self.row_index}, {transpiled_column_header}] = {self.type_corrected_new_value}')
        elif is_datetime_dtype(column_dtype):
            code.append(f'{self.df_name}.at[{self.row_index}, {transpiled_column_header}] = pd.to_datetime(\"{self.type_corrected_new_value}\")')
        elif is_timedelta_dtype(column_dtype):
            code.append(f'{self.df_name}.at[{self.row_index}, {transpiled_column_header}] = pd.to_timedelta(\"{self.type_corrected_new_value}\")')
        else:
            code.append(f'{self.df_name}.at[{self.row_index}, {transpiled_column_header}] = \"{self.type_corrected_new_value}\"')

        return code, [] # TODO: we might need pandas here as well!

    def get_edited_sheet_indexes(self) -> List[int]:
        return [self.sheet_index]

    def _combine_right_with_delete_column_code_chunk(self, other_code_chunk: DeleteColumnsCodeChunk) -> Optional["CodeChunk"]:
        return get_right_combine_with_column_delete_code_chunk(
            self,
            other_code_chunk,
            'sheet_index',
            'column_id',
        )
    def combine_right(self, other_code_chunk: CodeChunk) -> Optional[CodeChunk]:
        if isinstance(other_code_chunk, DeleteColumnsCodeChunk):
            return self._combine_right_with_delete_column_code_chunk(other_code_chunk)

        return None


#!/usr/bin/env python
# coding: utf-8

# Copyright (c) Saga Inc.
# Distributed under the terms of the GPL License.
import csv
import json
from os.path import normpath, basename
import os
from copy import copy
from time import perf_counter
from typing import Any, Dict, List, Optional, Set, Tuple
import chardet
import pandas as pd
from mitosheet.code_chunks.code_chunk import CodeChunk
from mitosheet.code_chunks.step_performers.import_steps.simple_import_code_chunk import DEFAULT_DECIMAL, DEFAULT_DELIMETER, DEFAULT_ENCODING, DEFAULT_ERROR_BAD_LINES, DEFAULT_SKIPROWS, SimpleImportCodeChunk, get_read_csv_params
from mitosheet.step_performers.utils import get_param

from mitosheet.utils import get_valid_dataframe_names
from mitosheet.errors import make_file_not_found_error, make_invalid_simple_import_error, make_is_directory_error
from mitosheet.state import DATAFRAME_SOURCE_IMPORTED, State
from mitosheet.step_performers.step_performer import StepPerformer


class SimpleImportStepPerformer(StepPerformer):
    """
    A simple import, which allows you to import csv files 
    with the given file_names, while detecting the correct
    way to import them.
    """

    @classmethod
    def step_version(cls) -> int:
        return 2

    @classmethod
    def step_type(cls) -> str:
        return 'simple_import'

    @classmethod
    def execute(cls, prev_state: State, params: Dict[str, Any]) -> Tuple[State, Optional[Dict[str, Any]]]:
        file_names: List[str] = get_param(params, 'file_names')
        delimeters: Optional[List[str]] = get_param(params, 'delimeters')
        encodings: Optional[List[str]] = get_param(params, 'encodings')
        decimals: Optional[List[str]] = get_param(params, 'decimals')
        skiprows: Optional[List[int]] = get_param(params, 'skiprows')
        error_bad_lines: Optional[List[bool]] = get_param(params, 'error_bad_lines')
        use_deprecated_id_algorithm: bool = get_param(params, 'use_deprecated_id_algorithm') if get_param(params, 'use_deprecated_id_algorithm') else False

        # If any of the files are directories, we throw an error to let
        # the user know
        for file_name in file_names:
            if os.path.isdir(file_name):
                raise make_is_directory_error(file_name)

        # Create a new step
        post_state = prev_state.copy()

        file_delimeters = []
        file_encodings = []
        file_decimals = []
        file_skiprows = []
        file_error_bad_lines = []

        just_final_file_names = [basename(normpath(file_name)) for file_name in file_names]

        pandas_processing_time = 0.0
        for index, (file_name, df_name) in enumerate(zip(file_names, get_valid_dataframe_names(post_state.df_names, just_final_file_names))):
            
            partial_pandas_start_time = perf_counter()

            try:
                # We try to read the csv with the parameters that the user specified. 
                # If the user has not specified parameters, then its because they did not go to the csv configure page, and instead
                # are using Mito's defaults. In this case, we're able to make educated guesses for the delimiter and encoding, and use pandas defaults
                # for the remainder of the parameters. 
                if delimeters is not None and encodings is not None:
                    delimeter = delimeters[index]
                    encoding = encodings[index]
                    
                    # Given the Mito UI, we expect that if the user has specified the delimiter and ecoding, 
                    # that the rest of the parameters are also defined. The only time that is not the case is when 
                    # the user is replaying an old simple import that does not have these fields. To account for that, 
                    # we just handle the None case here. This makes it easy to add new parameters without having to write 
                    # step upgraders. 
                    # This approach of handling optional step params instead of writing a step upgrader is also used in graphs.
                    decimal = decimals[index] if decimals is not None else DEFAULT_DECIMAL
                    _skiprows = skiprows[index] if skiprows is not None else DEFAULT_SKIPROWS
                    _error_bad_lines = error_bad_lines[index] if error_bad_lines is not None else DEFAULT_ERROR_BAD_LINES
                    df = pd.read_csv(file_name, **get_read_csv_params(delimeter, encoding, decimal, _skiprows, _error_bad_lines))
                    pandas_processing_time += (perf_counter() - partial_pandas_start_time)
                else:
                    # If the user does not specify the delimiter and encoding, then we guess them and use default values for everything else.
                    df, delimeter, encoding = read_csv_get_delimiter_and_encoding(file_name)
                    decimal = DEFAULT_DECIMAL
                    _skiprows = DEFAULT_SKIPROWS
                    _error_bad_lines = DEFAULT_ERROR_BAD_LINES
                    pandas_processing_time += (perf_counter() - partial_pandas_start_time)
            except:
                if os.path.exists(file_name):
                    raise make_invalid_simple_import_error()
                else:
                    raise make_file_not_found_error(file_name)

            # Save the delimeter and encodings for transpiling
            file_delimeters.append(delimeter)
            file_encodings.append(encoding)
            file_decimals.append(decimal)
            file_skiprows.append(_skiprows)
            file_error_bad_lines.append(_error_bad_lines)
            
            post_state.add_df_to_state(
                df, 
                DATAFRAME_SOURCE_IMPORTED, 
                df_name=df_name,
                use_deprecated_id_algorithm=use_deprecated_id_algorithm
            )   

        # Save the renames that have occured in the step, for transpilation reasons
        # and also save the seperator that we used for each file
        return post_state, {
            'file_delimeters': file_delimeters,
            'file_encodings': file_encodings,
            'file_decimals': file_decimals,
            'file_skiprows': file_skiprows,
            'file_error_bad_lines': file_error_bad_lines,
            'pandas_processing_time': pandas_processing_time
        }

    @classmethod
    def transpile(
        cls,
        prev_state: State,
        post_state: State,
        params: Dict[str, Any],
        execution_data: Optional[Dict[str, Any]],
    ) -> List[CodeChunk]:
        return [
            SimpleImportCodeChunk(
                prev_state, 
                post_state, 
                get_param(params, 'file_names'), 
                get_param(execution_data if execution_data is not None else {}, 'file_delimeters'), 
                get_param(execution_data if execution_data is not None else {}, 'file_encodings'), 
                get_param(execution_data if execution_data is not None else {}, 'file_decimals'), 
                get_param(execution_data if execution_data is not None else {}, 'file_skiprows'), 
                get_param(execution_data if execution_data is not None else {}, 'file_error_bad_lines')
            )
        ]
    
    @classmethod
    def get_modified_dataframe_indexes(cls, params: Dict[str, Any]) -> Set[int]:
        return {-1}




def read_csv_get_delimiter_and_encoding(file_name: str) -> Tuple[pd.DataFrame, str, str]:
    """
    Given a file_name, will read in the file as a CSV, and
    return the df, delimeter, decimal separator, and encoding of the file
    """
    encoding = DEFAULT_ENCODING
    delimeter = DEFAULT_DELIMETER
    try:
        # First attempt to read csv without specifying an encoding, just with a delimeter
        delimeter = guess_delimeter(file_name)
        df = pd.read_csv(file_name, sep=delimeter)
    except UnicodeDecodeError:
        # If we have an encoding error, try and get the encoding
        try: 
            encoding = guess_encoding(file_name)
            delimeter = guess_delimeter(file_name, encoding=encoding)

            # Read the file as dataframe 
            df = pd.read_csv(file_name, sep=delimeter, encoding=encoding)
        except: 
            # Sometimes guess_encoding, guesses 'ascii' when we want 'latin-1', 
            # so if guess_encoding fails, we try latin-1
            encoding = 'latin-1'
            df = pd.read_csv(file_name, sep=delimeter, encoding=encoding)
        
    return df, delimeter, encoding


def guess_delimeter(file_name: str, encoding: Optional[str]=None) -> str:
    """
    Given a path to a file that is assumed to exist and be a CSV, this
    function guesses the delimeter that is used by that file
    """
    s = csv.Sniffer()
    with open(file_name, 'r', encoding=encoding) as f:
        return s.sniff(f.readline()).delimiter

def guess_encoding(file_name: str) -> str:
    """
    Uses chardet to guess the encoding of the the file
    at the given file_name
    """
    # Attempt to determine the encoding and try again. 
    with open(file_name, 'rb') as f:
        result = chardet.detect(f.readline())
        return result['encoding']






import string
from typing import Optional, Tuple, Union

from openpyxl import load_workbook

from mitosheet.errors import make_invalid_range_error


def get_excel_range_from_column_index(col_index: int) -> str:
    """
    Number to Excel-style column name, e.g., 1 = A:A, 26 = Z:Z, 27 = AA:AA, 703 = AAA:AAA.
    """
    return get_column_from_column_index(col_index) + ":" + get_column_from_column_index(col_index)

def get_column_from_column_index(col_index: int) -> str:
    """
    Number to Excel-style column name, e.g., 1 = A, 26 = Z, 27 = AA, 703 = AAA.
    """
    # Add 1 because Mito 0 indexes columns
    col_index = col_index + 1
    name = ''
    while col_index > 0:
        col_index, r = divmod (col_index - 1, 26)
        name = chr(r + ord('A')) + name
    return name

def get_col_and_row_indexes_from_range(range: str) -> Tuple[Tuple[int, int], Tuple[int, int]]:
    try:
        start_cell_address, end_cell_address = range.split(':')
        (start_col_idx, start_row_idx) = get_row_and_col_index_from_cell_address(start_cell_address)
        (end_col_idx, end_row_idx) = get_row_and_col_index_from_cell_address(end_cell_address)
    except:
        raise make_invalid_range_error(range, False)

    return ((min(start_col_idx, end_col_idx), min(start_row_idx, end_row_idx)), (max(start_col_idx, end_col_idx), max(start_row_idx, end_row_idx)))

def get_row_and_col_index_from_cell_address(cell_address: str) -> Tuple[int, int]:
    
    letters = ''
    numbers = ''
    for char in cell_address:
        if char.isalpha():
            letters += char
        else:
            numbers += char

    return (get_index_from_column(letters), int(numbers) - 1)

def get_index_from_column(col: str) -> int:
    num = 0
    for c in col:
        if c in string.ascii_letters:
            num = num * 26 + (ord(c.upper()) - ord('A')) + 1
    return num - 1


def get_table_range_from_upper_left_corner_value(file_path: str, sheet_name: str, value: Union[str, int, float, bool]) -> Optional[str]:
    """
    Given a string, this function will look through the excel tab sheet_name at the given
    file_path, and find the first instance of the value (look through column A first, then B, etc).

    If this value does not exist in the tab, will return None. 

    If the value does exist, then this function will walk down the column until it hits an empty row. Then,
    it will walk down the final row until it hits an empty column. Then it will return the range that 
    defines that rectangular of defined data (with value in the upper left corner).
    """
    workbook = load_workbook(file_path)
    sheet = workbook[sheet_name]

    # We get the last defined rows, so we don't waste time searching data we don't need
    dimension = sheet.calculate_dimension()
    (min_search_col, min_search_row), (max_search_col, max_search_row) = get_col_and_row_indexes_from_range(dimension)
    # Unfortunately, openpyxl indexes from 1, so we add one to treat everything in that range
    (min_search_col, min_search_row), (max_search_col, max_search_row) = (min_search_col + 1, min_search_row + 1), (max_search_col + 1, max_search_row + 1)

    # Loop over the columns one by one to find where this value is set
    min_found_col_index, min_found_row_index = None, None
    for col in sheet.iter_cols(min_row=min_search_row, max_row=max_search_row, min_col=min_search_col, max_col=max_search_col):
        for cell in col:
            if cell.value == value:
                min_found_col_index, min_found_row_index = cell.column, cell.row
                break
        
        # As soon as we find something, stop looking
        if min_found_col_index is not None or min_found_row_index is not None:
            break

    if min_found_col_index is None or min_found_row_index is None:
        return None

    column = sheet[get_column_from_column_index(min_found_col_index - 1)] # We need to subtract 1 as we 0 index
    max_found_row_index = None
    for cell in column:
        if cell.row < min_found_row_index:
            continue
        
        if cell.value is None:
            max_found_row_index = cell.row - 1 # minus b/c this is one past the end
            break

    # If we looped over the entire column without ending, then we set the max row index
    # as the length of the entire column
    if max_found_row_index is None:
        max_found_row_index = len(column)

    # Then we find find where the rows are defined to
    max_found_col_index = None
    for row in sheet.iter_rows(min_row=max_found_row_index - 1, max_row=max_found_row_index - 1, min_col=min_found_col_index):
        for cell in row:
            if cell.value is None:
                max_found_col_index = cell.column - 1 # minus b/c this is one past the end
                break
    
    # Similarly, if we don't find any empty value in the defined cells, we set the max_col index
    # as the limit of the sheet
    if max_found_col_index is None:
        max_found_col_index = max_search_col

    return f'{get_column_from_column_index(min_found_col_index - 1)}{min_found_row_index}:{get_column_from_column_index(max_found_col_index - 1)}{max_found_row_index}'


def get_df_name_as_valid_sheet_name(df_name: str) -> str:
    return df_name[:31] # not more than 32 chars


#!/usr/bin/env python
# coding: utf-8

# Copyright (c) Saga Inc.
# Distributed under the terms of the GPL License.
from copy import deepcopy
from time import perf_counter
from typing import Any, Dict, List, Optional, Set, Tuple

import pandas as pd
from mitosheet.code_chunks.code_chunk import CodeChunk
from mitosheet.code_chunks.step_performers.column_steps.reorder_column_code_chunk import ReorderColumnCodeChunk
from mitosheet.state import State
from mitosheet.step_performers.step_performer import StepPerformer
from mitosheet.step_performers.utils import get_param
from mitosheet.types import ColumnHeader, ColumnID


def get_valid_index(dfs: List[pd.DataFrame], sheet_index: int, new_column_index: int) -> int:
    # make sure new_column_index is valid
    if new_column_index < 0:
        new_column_index = 0

    if new_column_index >= len(dfs[sheet_index].columns):
        new_column_index = len(dfs[sheet_index].columns) - 1

    return new_column_index


class ReorderColumnStepPerformer(StepPerformer):
    """""
    A reorder_column step, which allows you to move 
    a column to a different location in the df.
    """

    @classmethod
    def step_version(cls) -> int:
        return 2

    @classmethod
    def step_type(cls) -> str:
        return 'reorder_column'

    @classmethod
    def execute(cls, prev_state: State, params: Dict[str, Any]) -> Tuple[State, Optional[Dict[str, Any]]]:
        sheet_index: int = get_param(params, 'sheet_index')
        column_id: ColumnID = get_param(params, 'column_id')
        new_column_index: int = get_param(params, 'new_column_index')

        column_header = prev_state.column_ids.get_column_header_by_id(sheet_index, column_id)

        new_column_index = get_valid_index(prev_state.dfs, sheet_index, new_column_index)
            
        # Create a new post state
        post_state = prev_state.copy(deep_sheet_indexes=[sheet_index])

        # Actually execute the column reordering
        pandas_start_time = perf_counter()
        final_df = _execute_reorder_column(
            prev_state.dfs[sheet_index],
            column_header,
            new_column_index
        )
        pandas_processing_time = perf_counter() - pandas_start_time

        post_state.dfs[sheet_index] = final_df

        return post_state, {
            'pandas_processing_time': pandas_processing_time
        }

    @classmethod
    def transpile(
        cls,
        prev_state: State,
        post_state: State,
        params: Dict[str, Any],
        execution_data: Optional[Dict[str, Any]],
    ) -> List[CodeChunk]:
        return [
            ReorderColumnCodeChunk(
                prev_state, 
                post_state, 
                get_param(params, 'sheet_index'),
                get_param(params, 'column_id'),
                get_param(params, 'new_column_index'),
            )
        ]
    
    @classmethod
    def get_modified_dataframe_indexes(cls, params: Dict[str, Any]) -> Set[int]:
        return {get_param(params, 'sheet_index')}


def _execute_reorder_column(df: pd.DataFrame, column_header: ColumnHeader, new_column_index: int) -> pd.DataFrame:
    """
    Helper function for reordering a column in the dataframe
    """
    df_columns = [col for col in df.columns if col != column_header]
    df_columns.insert(new_column_index, column_header)
    return df[df_columns]



#!/usr/bin/env python
# coding: utf-8

# Copyright (c) Saga Inc.
# Distributed under the terms of the GPL License.
from typing import Any, Dict, List
from mitosheet.saved_analyses.step_upgraders.utils_column_header_to_column_id import \
    replace_headers_with_id
from mitosheet.saved_analyses.step_upgraders.utils_rename_column_headers import \
    BULK_OLD_RENAME_STEP
from mitosheet.step_performers.pivot import PCT_NO_OP


def upgrade_group_1_to_pivot_2(step: Dict[str, Any], later_steps: List[Dict[str, Any]]) -> List[Dict[str, Any]]:
    """
    Upgrades from a group 1 step to a pivot 2 step, simply
    by changing the names of the params.

    Old format of the step: {
        "step_version": 1, 
        "step_type": "group", 
        "sheet_index": x, 
        "group_rows": [...], 
        "group_columns": [...], 
        "values": {...: ...}}}
    }

    New format of the step: {
        "step_version": 2, 
        "step_type": "pivot", 
        "sheet_index": old['sheet_index'], 
        "pivot_rows": old['group_rows'], 
        "pivot_columns": old['group_columns'], 
        "values": old['values']
    }
    """
    return [{
        'step_version': 2,
        'step_type': 'pivot',
        'sheet_index': step['sheet_index'],
        'pivot_rows': step['group_rows'],
        'pivot_columns': step['group_columns'],
        'values': step['values']
    }] + later_steps


def upgrade_pivot_2_to_pivot_3(step: Dict[str, Any], later_steps: List[Dict[str, Any]]) -> List[Dict[str, Any]]:
    """
    Upgrades from a pivot 2 step to a pivot 3 step, which simple
    changes the formats of the values key to be a list rather than
    a single element

    Old format of the step: {
        "step_version": 2, 
        "step_type": "pivot", 
        "sheet_index": int, 
        "group_rows": string[], 
        "group_columns": string[], 
        "values": {string: AggregationType}
    }

    New format of the step: {
        "step_version": 3, 
        "step_type": "pivot", 
        "sheet_index": old['sheet_index'], 
        "pivot_rows": old['pivot_rows'], 
        "pivot_columns": old['pivot_columns'], 
        "values": {string: AggregationType[]} <- note the single item in this list is the previous single value
    }
    """
    return [{
        'step_version': 3,
        'step_type': 'pivot',
        'sheet_index': step['sheet_index'],
        'pivot_rows': step['pivot_rows'],
        'pivot_columns': step['pivot_columns'],
        'values': {
            column_header: [aggregationType] for column_header, aggregationType in step['values'].items()
        }
    }] + later_steps


def upgrade_pivot_3_to_4(step: Dict[str, Any], later_steps: List[Dict[str, Any]]) -> List[Dict[str, Any]]:
    """
    Moves to using column id instead of column header.

    OLD: {
        "step_version": 3, 
        "step_type": "pivot", 
        "params": {
            sheet_index: 0,
            pivot_rows: List[_header_],
            pivot_columns: List[_header_],
            values: Dict[_header_, List[str]],
            destination_sheet_index: 1,
        }
    }

    NEW: {
        "step_version": 3, 
        "step_type": "pivot", 
        "params": {
            sheet_index: 0,
            pivot_rows_column_ids: List[_id_],
            pivot_columns_column_ids: List[_id_],
            values_column_ids_map: Dict[_id_, List[str]],
            destination_sheet_index: 1,
        }
    }
    """
    params = step['params']
    params = replace_headers_with_id(params, 'pivot_rows', 'pivot_rows_column_ids')
    params = replace_headers_with_id(params, 'pivot_columns', 'pivot_columns_column_ids')
    params = replace_headers_with_id(params, 'values', 'values_column_ids_map')

    return [{
        "step_version": 4, 
        "step_type": "pivot", 
        "params": params
    }] + later_steps


def upgrade_pivot_4_to_5_and_rename(step: Dict[str, Any], later_steps: List[Dict[str, Any]]) -> List[Dict[str, Any]]:
    """
    Adds a rename step after pivot 4, as in the move from pivot 4 -> 5
    we got rid of the column headers that we used. Also makes sure that
    the old id generation algorithm is used.

    Old format: {
        "step_version": 4, 
        "step_type": "pivot", 
        "params": {
            sheet_index: 0,
            pivot_rows_column_ids: List[_id_],
            pivot_columns_column_ids: List[_id_],
            values_column_ids_map: Dict[_id_, List[str]],
            destination_sheet_index: 1,
        }
    }

    New format: [
        {
            "step_version": 5, 
            "step_type": "pivot", 
            "params": {
                sheet_index: 0,
                pivot_rows_column_ids: List[_id_],
                pivot_columns_column_ids: List[_id_],
                values_column_ids_map: Dict[_id_, List[str]],
                destination_sheet_index: 1,
                use_deprecated_id_algorithm: true
            }
        },
        {
            "step_version": 1, 
            "step_type": "bulk_old_rename", 
            "params": {}
        },
    ]    
    """

    step['step_version'] = 5
    step['params']['use_deprecated_id_algorithm'] = True
    
    return [
        step, 
        BULK_OLD_RENAME_STEP
    ] + later_steps


def upgrade_pivot_5_to_6(step: Dict[str, Any], later_steps: List[Dict[str, Any]]) -> List[Dict[str, Any]]:
    """
    When moving from pivot 5 to pivot 6, we got rid of the flattening that
    happened with column headers. As such, we add a parameter to the pivot
    step to tell it to flatten the headers.

    Old format: [
        {
            "step_version": 5, 
            "step_type": "pivot", 
            "params": {
                sheet_index: 0,
                pivot_rows_column_ids: List[_id_],
                pivot_columns_column_ids: List[_id_],
                values_column_ids_map: Dict[_id_, List[str]],
                destination_sheet_index: 1,
            }
        },
    ]

    New format: [
        {
            "step_version": 6, 
            "step_type": "pivot", 
            "params": {
                sheet_index: 0,
                pivot_rows_column_ids: List[_id_],
                pivot_columns_column_ids: List[_id_],
                values_column_ids_map: Dict[_id_, List[str]],
                destination_sheet_index: 1,
                flatten_column_headers: True
            }
        },
    ]
    """
    step['step_version'] = 6
    step['params']['flatten_column_headers'] = True
    
    return [step] + later_steps

def upgrade_pivot_6_to_7(step: Dict[str, Any], later_steps: List[Dict[str, Any]]) -> List[Dict[str, Any]]:
    """
    When moving from pivot 6 to 7, we added the pivot_filters key, which contains
    a list of filters. Each filter is applied to a specific column.

    Old format: [
        {
            "step_version": 6, 
            "step_type": "pivot", 
            "params": {
                sheet_index: 0,
                pivot_rows_column_ids: List[_id_],
                pivot_columns_column_ids: List[_id_],
                values_column_ids_map: Dict[_id_, List[str]],
                destination_sheet_index: 1,
                flatten_column_headers: True
            }
        },
    ]

    New format: [
        {
            "step_version": 7, 
            "step_type": "pivot", 
            "params": {
                sheet_index: 0,
                pivot_rows_column_ids: List[_id_],
                pivot_columns_column_ids: List[_id_],
                values_column_ids_map: Dict[_id_, List[str]],
                pivot_filters: {columnID: ColumnID, filter: FilterType}[]
                destination_sheet_index: 1,
                flatten_column_headers: True
            }
        },
    ]
    """
    step['step_version'] = 7
    step['params']['pivot_filters'] = []
    
    return [step] + later_steps

def upgrade_pivot_7_to_8(step: Dict[str, Any], later_steps: List[Dict[str, Any]]) -> List[Dict[str, Any]]:
    """
    When moving from pivot 7 to 8, we change the type of pivot_rows_column_ids and pivot_columns_column_ids
    from ColumnID[] => {column_id: ColumnID, transformation: PivotColumnTransformation}[]. For now, all of
    them are defaulted to `no-op`. It also renames their keys to pivot_rows_column_ids_with_transforms and 
    pivot_columns_column_ids_with_transforms to be more clear what they are.

    Old format: [
        {
            "step_version": 7, 
            "step_type": "pivot", 
            "params": {
                sheet_index: 0,
                pivot_rows_column_ids: List[_id_],
                pivot_columns_column_ids: List[_id_],
                values_column_ids_map: Dict[_id_, List[str]],
                pivot_filters: {columnID: ColumnID, filter: FilterType}[]
                destination_sheet_index: 1,
                flatten_column_headers: True
            }
        },
    ]
    Old format: [
        {
            "step_version": 8, 
            "step_type": "pivot", 
            "params": {
                sheet_index: 0,
                pivot_rows_column_ids_with_transforms: {column_id: ColumnID, transformation: PivotColumnTransformation}[],
                pivot_columns_column_ids_with_transforms: {column_id: ColumnID, transformation: PivotColumnTransformation}[],
                values_column_ids_map: Dict[_id_, List[str]],
                pivot_filters: {columnID: ColumnID, filter: FilterType}[]
                destination_sheet_index: 1,
                flatten_column_headers: True
            }
        },
    ]
    """
    step['step_version'] = 8
    step['params']['pivot_rows_column_ids_with_transforms'] = [{
        'column_id': column_id, 'transformation': PCT_NO_OP 
    } for column_id in step['params']['pivot_rows_column_ids']]
    step['params']['pivot_columns_column_ids_with_transforms'] = [{
        'column_id': column_id, 'transformation': PCT_NO_OP
    } for column_id in step['params']['pivot_columns_column_ids']]

    del step['params']['pivot_rows_column_ids']
    del step['params']['pivot_columns_column_ids']
    
    return [step] + later_steps




#!/usr/bin/env python
# coding: utf-8

# Copyright (c) Saga Inc.
# Distributed under the terms of the GPL License.

from copy import copy
from time import perf_counter
from typing import Any, Dict, List, Optional, Set, Tuple

import pandas as pd
from mitosheet.code_chunks.code_chunk import CodeChunk
from mitosheet.code_chunks.step_performers.merge_code_chunk import MergeCodeChunk
from mitosheet.errors import (get_recent_traceback, make_incompatible_merge_headers_error,
                              make_incompatible_merge_key_error)
from mitosheet.state import DATAFRAME_SOURCE_MERGED, State
from mitosheet.step_performers.step_performer import StepPerformer
from mitosheet.step_performers.utils import get_param
from mitosheet.transpiler.transpile_utils import (
    column_header_list_to_transpiled_code, column_header_to_transpiled_code)
from mitosheet.types import ColumnHeader, ColumnID

LOOKUP = 'lookup'
UNIQUE_IN_LEFT = 'unique in left'
UNIQUE_IN_RIGHT = 'unique in right'

class MergeStepPerformer(StepPerformer):
    """
    Allows you to merge two dataframes together.
    """

    @classmethod
    def step_version(cls) -> int:
        return 4

    @classmethod
    def step_type(cls) -> str:
        return 'merge'

    @classmethod
    def execute(cls, prev_state: State, params: Dict[str, Any]) -> Tuple[State, Optional[Dict[str, Any]]]:
        how: str = get_param(params, 'how')
        sheet_index_one: int = get_param(params, 'sheet_index_one')
        sheet_index_two: int = get_param(params, 'sheet_index_two')
        merge_key_column_ids: List[List[ColumnID]] = get_param(params, 'merge_key_column_ids')
        selected_column_ids_one: List[ColumnID] = get_param(params, 'selected_column_ids_one')
        selected_column_ids_two: List[ColumnID] = get_param(params, 'selected_column_ids_two')

        merge_keys_one = prev_state.column_ids.get_column_headers_by_ids(sheet_index_one, list(map(lambda x: x[0], merge_key_column_ids)))
        merge_keys_two = prev_state.column_ids.get_column_headers_by_ids(sheet_index_two, list(map(lambda x: x[1], merge_key_column_ids)))

        selected_columns_one = prev_state.column_ids.get_column_headers_by_ids(sheet_index_one, selected_column_ids_one)
        selected_columns_two = prev_state.column_ids.get_column_headers_by_ids(sheet_index_two, selected_column_ids_two)

        # We create a shallow copy to make the new post state
        post_state = prev_state.copy()

        pandas_start_time = perf_counter()
        new_df = _execute_merge(
            prev_state.dfs,
            prev_state.df_names,
            how,
            sheet_index_one,
            merge_keys_one,
            selected_columns_one,
            sheet_index_two,
            merge_keys_two,
            selected_columns_two
        )
        pandas_processing_time = perf_counter() - pandas_start_time

        # Add this dataframe to the new post state
        post_state.add_df_to_state(new_df, DATAFRAME_SOURCE_MERGED)

        return post_state, {
            'pandas_processing_time': pandas_processing_time
        }

    @classmethod
    def transpile(
        cls,
        prev_state: State,
        post_state: State,
        params: Dict[str, Any],
        execution_data: Optional[Dict[str, Any]],
    ) -> List[CodeChunk]:
        return [
            MergeCodeChunk(
                prev_state, 
                post_state, 
                get_param(params, 'how'),
                get_param(params, 'sheet_index_one'),
                get_param(params, 'sheet_index_two'),
                get_param(params, 'merge_key_column_ids'),
                get_param(params, 'selected_column_ids_one'),
                get_param(params, 'selected_column_ids_two'),
            )
        ]
    
    @classmethod
    def get_modified_dataframe_indexes(cls, params: Dict[str, Any]) -> Set[int]:
        return {-1}

def _execute_merge(
        dfs: List[pd.DataFrame], 
        df_names: List[str],
        how: str,
        sheet_index_one: int,
        merge_keys_one: List[ColumnHeader], 
        selected_columns_one: List[ColumnHeader],
        sheet_index_two: int,
        merge_keys_two: List[ColumnHeader],
        selected_columns_two: List[ColumnHeader]
    ) -> pd.DataFrame:
    """
    Executes a merge on the sheets with the given indexes, merging on the 
    given keys, and only keeping the selection columns from each df.
    """
    # We currently error if you try and merge two dataframes where they
    # have different levels to the multi-index in the column headers. Aka, 
    # you cannot merge a multi-index dataframe into a regular dataframe
    if dfs[sheet_index_one].columns.nlevels != dfs[sheet_index_two].columns.nlevels:
        raise make_incompatible_merge_headers_error(error_modal=False)

    # If there's no merge keys, we return an empty dataframe
    if len(merge_keys_one) == 0 and len(merge_keys_two) == 0:
        return pd.DataFrame()

    if how == LOOKUP:
        # We drop duplicates to avoid pairwise duplication on the merge.
        temp_df = dfs[sheet_index_two].drop_duplicates(subset=merge_keys_two)
        # We overwrite the how variable to 'left' so it can be used in the merge
        how_to_use = 'left'
    else:
        temp_df = dfs[sheet_index_two]
        how_to_use = how

    # Then we delete all the columns from each we don't wanna keep (making sure to keep the merge keys)
    deleted_columns_one = set(dfs[sheet_index_one].keys()).difference(set(selected_columns_one).union(set(merge_keys_one)))
    deleted_columns_two = set(dfs[sheet_index_two].keys()).difference(set(selected_columns_two).union(set(merge_keys_two)))

    df_one_cleaned = dfs[sheet_index_one].drop(deleted_columns_one, axis=1)
    df_two_cleaned = temp_df.drop(deleted_columns_two, axis=1)

    # Finially, we perform the merge!
    df_one_name = df_names[sheet_index_one]
    df_two_name = df_names[sheet_index_two]
    # We make sure the suffixes aren't the same, as otherwise we might end up with 
    # one df with duplicated column headers
    suffix_one = df_one_name
    suffix_two = df_two_name if df_two_name != df_one_name else f'{df_two_name}_2'

    try:
        if how == UNIQUE_IN_LEFT:
            df_two_cleaned = df_two_cleaned.drop_duplicates(subset=merge_keys_two) # Remove duplicates so lookup merge only returns first match
            # Create a boolean index array for values only in the left sheet by turning on the indicator=True and then filtering on the result
            bool_index_array = df_one_cleaned.merge(df_two_cleaned, left_on=merge_keys_one, right_on=merge_keys_two, how='left', indicator=True)['_merge'] == 'left_only'
            # Get the final result by filtering on the boolean index array and then selecting the columns we want to keep
            return df_one_cleaned.copy(deep=True)[bool_index_array][selected_columns_one].reset_index(drop=True)
        if how == UNIQUE_IN_RIGHT:
            df_one_cleaned = df_one_cleaned.drop_duplicates(subset=merge_keys_one) # Remove duplicates so lookup merge only returns first match
            bool_index_array = df_one_cleaned.merge(df_two_cleaned, left_on=merge_keys_one, right_on=merge_keys_two, how='right', indicator=True)['_merge'] == 'right_only'
            return df_two_cleaned.copy(deep=True)[bool_index_array][selected_columns_two].reset_index(drop=True)
        else:
            return df_one_cleaned.merge(df_two_cleaned, left_on=merge_keys_one, right_on=merge_keys_two, how=how_to_use, suffixes=[f'_{suffix_one}', f'_{suffix_two}'])
    except ValueError:
        # If we get a value error from merging two incompatible columns, we go through and check 
        # to see which of the columns this is, so our error can be maximally informative
        for merge_key_one, merge_key_two in zip(merge_keys_one, merge_keys_two):
            merge_key_one_dtype = str(dfs[sheet_index_one][merge_key_one].dtype)
            merge_key_two_dtype = str(dfs[sheet_index_two][merge_key_two].dtype)

            if merge_key_one_dtype != merge_key_two_dtype:
                raise make_incompatible_merge_key_error(
                    merge_key_one=merge_key_one, 
                    merge_key_one_dtype=merge_key_one_dtype,
                    merge_key_two=merge_key_two, 
                    merge_key_two_dtype=merge_key_two_dtype,
                    error_modal=False
                )

        raise make_incompatible_merge_key_error(error_modal=False)





#!/usr/bin/env python
# coding: utf-8

# Copyright (c) Saga Inc.
# Distributed under the terms of the GPL License.

from time import perf_counter
from typing import Any, Dict, List, Optional, Set, Tuple
from mitosheet.code_chunks.code_chunk import CodeChunk
from mitosheet.code_chunks.reset_index_code_chunk import ResetIndexCodeChunk

from mitosheet.state import State
from mitosheet.step_performers.step_performer import StepPerformer
from mitosheet.step_performers.utils import get_param

class ResetIndexStepPerformer(StepPerformer):
    """
    Allows you to reset index.
    """

    @classmethod
    def step_version(cls) -> int:
        return 1

    @classmethod
    def step_type(cls) -> str:
        return 'reset_index'

    @classmethod
    def execute(cls, prev_state: State, params: Dict[str, Any]) -> Tuple[State, Optional[Dict[str, Any]]]:
        sheet_index: int = get_param(params, 'sheet_index')
        drop: bool = get_param(params, 'drop')

        # We make a new state to modify it
        post_state = prev_state.copy(deep_sheet_indexes=[sheet_index]) 

        pandas_start_time = perf_counter()

        post_state.dfs[sheet_index] = post_state.dfs[sheet_index].reset_index(drop=drop)
        if not drop:
            # If we keep the index as a column, keep track of it's metadata (it gets added as the first column)
            # TODO: there's a bug if you don't drop multiple times, eventually it errors
            post_state.add_columns_to_state(sheet_index, [post_state.dfs[sheet_index].columns[0]]) 
        
        pandas_processing_time = perf_counter() - pandas_start_time


        return post_state, {
            'pandas_processing_time': pandas_processing_time,
            'result': {
                # TODO: fill in the result
            }
        }

    @classmethod
    def transpile(
        cls,
        prev_state: State,
        post_state: State,
        params: Dict[str, Any],
        execution_data: Optional[Dict[str, Any]],
    ) -> List[CodeChunk]:
        return [
            ResetIndexCodeChunk(
                prev_state, 
                post_state, 
                get_param(params, 'sheet_index'),
                get_param(params, 'drop'),
            )
        ]

    @classmethod
    def get_modified_dataframe_indexes(cls, params: Dict[str, Any]) -> Set[int]:
        return {get_param(params, 'sheet_index')}
    


"""
Utilities for figuring out where Mito is being run.
"""

import os

from mitosheet.utils import run_command


def is_in_google_colab() -> bool:
    # From: https://discourse.jupyter.org/t/find-out-if-my-code-runs-inside-a-notebook-or-jupyter-lab/6935/19
    try:
        from IPython.core.getipython import get_ipython
        return 'google.colab' in str(get_ipython()) # type: ignore
    except:
        return False

def is_in_vs_code() -> bool:
    # From: https://github.com/microsoft/vscode-jupyter/issues/3364
    return 'VSCODE_PID' in os.environ

def is_jupyter_lab_running() -> bool:
    """
    Format of command output: 

    Currently running servers:
    http://localhost:8888/?token=wenflwenflqnwdlkqmwldkm :: Path/to/launch
    """
    try:
        return '::' in run_command(['jupyter', 'lab', 'list'])[0]
    except:
        return False

def is_jupyter_notebook_running() -> bool:
    """
    Format of command output: 

    Currently running servers:
    http://localhost:8888/?token=wenflwenflqnwdlkqmwldkm :: Path/to/launch
    """
    try:
        return '::' in run_command(['jupyter', 'notebook', 'list'])[0]
    except:
        return False

def is_notebook() -> bool:
    """
    Returns true if this code is run in some sort of notebook
    Taken from: https://stackoverflow.com/questions/15411967/how-can-i-check-if-code-is-executed-in-the-ipython-notebook
    """
    try:
        from IPython.core.getipython import get_ipython
    except:
        return False
    try:
        shell = get_ipython().__class__.__name__ # type: ignore
        if shell == 'ZMQInteractiveShell':
            return True   # Jupyter notebook or qtconsole
        elif shell == 'TerminalInteractiveShell':
            return False  # Terminal running IPython
        else:
            return False  # Other type (?)
    except NameError:
        return False      # Probably standard Python interpreter

def get_location() -> str:
    notebook = is_notebook()
    lab_running = is_jupyter_lab_running()
    notebook_running = is_jupyter_notebook_running()

    if is_in_google_colab():
        return 'location_google_colab'
    elif is_in_vs_code():
        return 'location_vs_code'
    elif notebook and (lab_running and not notebook_running):
        return 'location_jupyter_lab'
    elif notebook and (not lab_running and notebook_running):
        return 'location_jupyter_notebook'
    elif notebook:
        # NOTE: in this case, both jlab and jnotebook are running, and
        # we cannot tell where we are. So we just say unknown
        return 'location_unknown_notebook'
    else:
        return 'location_unknown'


# https://stackoverflow.com/questions/43878953/how-does-one-detect-if-one-is-running-within-a-docker-container-within-python
def _is_docker() -> bool:
    path = '/proc/self/cgroup'
    return (
        os.path.exists('/.dockerenv') or
        os.path.isfile(path) and any('docker' in line for line in open(path))
    )
# We cache this so we don't have to take a performance impact
IS_DOCKER = _is_docker()
def is_docker() -> bool:
    return IS_DOCKER



#!/usr/bin/env python
# coding: utf-8

# Copyright (c) Saga Inc.
# Distributed under the terms of the GPL License.
"""
Contains all functions that don't particularly fit in any other category

All functions describe their behavior with a function documentation object
in the function docstring. Function documentation objects are described
in more detail in docs/README.md.

NOTE: This file is alphabetical order!
"""
import pandas as pd
import numpy as np
import datetime

from mitosheet.sheet_functions.sheet_function_utils import try_extend_series_to_index
from mitosheet.sheet_functions.types.decorators import filter_nans, convert_arg_to_series_type, handle_sheet_function_errors


@handle_sheet_function_errors
@convert_arg_to_series_type(
    0,
    'series',
    'error',
    'error'
)
@convert_arg_to_series_type(
    1,
    'series',
    'error',
    'error'
)
def FILLNAN(series: pd.Series, replacement: pd.Series) -> pd.Series:
    """
    {
        "function": "FILLNAN",
        "description": "Replaces the NaN values in the series with the replacement value.",
        "search_terms": ["fillnan", "nan", "fill nan", "missing values", "null", "null value", "fill null"],
        "examples": [
            "FILLNAN(A, 10)",
            "FILLNAN(A, 'replacement')"
        ],
        "syntax": "FILLNAN(series, replacement)",
        "syntax_elements": [{
                "element": "series",
                "description": "The series to replace the NaN values in."
            },
            {
                "element": "replacement",
                "description": "A string, number, or date to replace the NaNs with."
            }
        ]
    }
    """
    # Make sure the replacement series is long enough
    replacement = try_extend_series_to_index(replacement, series.index)

    return series.fillna(replacement)


@handle_sheet_function_errors
@convert_arg_to_series_type(
    0,
    'series',
    'error',
    'error'
)
def TYPE(series: pd.Series) -> pd.Series:
    """
    {
        "function": "TYPE",
        "description": "Returns the type of each element of the passed series. Return values are 'number', 'str', 'bool', 'datetime', 'object', or 'NaN'.",
        "search_terms": ["type", "dtype"],
        "examples": [
            "TYPE(Nums_and_Strings)",
            "IF(TYPE(Account_Numbers) != 'NaN', Account_Numbers, 0)"
        ],
        "syntax": "TYPE(series)",
        "syntax_elements": [{
                "element": "series",
                "description": "The series to get the type of each element of."
            }
        ]
    }
    """

    def get_element_type(element):
        try:
            # Try nan first, this may fail
            if np.isnan(element):
                return 'NaN'
        except:
            pass 

        # Start with bool!
        if isinstance(element, bool):
            return 'bool'
        elif isinstance(element, int):
            return 'number'
        elif isinstance(element, float):
            return 'number'
        elif isinstance(element, str):
            return 'string'
        elif isinstance(element, datetime.datetime):
            return 'datetime'
        elif isinstance(element, pd.Timedelta):
            return 'timedelta'
        return 'object'

    return series.apply(get_element_type).astype('str')


# TODO: we should see if we can list these automatically!
MISC_FUNCTIONS = {
    'FILLNAN': FILLNAN,
    'TYPE': TYPE
}


#!/usr/bin/env python
# coding: utf-8

# Copyright (c) Saga Inc.
# Distributed under the terms of the GPL License.

"""
Contains utilities for interacting with experiments. If you want to change the currently running
experiment, do the following:

1. Change the get_new_experiment() function to return the new experiment you want to run.
2. Change the get_new_experiment() in the mitoinstaller to use the new experiment id as well.
3. Add this to the experiment tracker in notion so we remember what the experiments are.
4. Actually use the new experiment (and remove old experiment code).

That's it! We can continue to optimize this over time, but that is fine for now.
"""


import json
import random
from typing import Dict, Optional

from mitosheet.user.db import USER_JSON_PATH, get_user_field

def get_random_variant() -> str:
    """Returns "A" or "B" with 50% probability
    
    Currently, as we have no experiment, only returns option B!
    """
    return "B"

def get_new_experiment() -> Dict[str, str]:
    # NOTE: this needs to match the installer!
    return {
        'experiment_id': 'installer_communication_and_time_to_value',
        'variant': get_random_variant(),
    }

def get_current_experiment() -> Optional[Dict[str, str]]:
    """
    Returns the current experiment object, or None if it does not exist
    """
    from mitosheet.user.schemas import UJ_EXPERIMENT

    try:
        return get_user_field(UJ_EXPERIMENT)
    except:
        return None

def set_experiment(experiment_id: str, variant: str) -> None:
    """
    Updates the running experiment and variant of the experiment with the given experiment_id 
    and variant.
    """
    from mitosheet.user.schemas import UJ_EXPERIMENT

    with open(USER_JSON_PATH, 'r') as user_file_old:
        old_user_json = json.load(user_file_old)
        old_user_json[UJ_EXPERIMENT]['experiment_id'] = experiment_id
        old_user_json[UJ_EXPERIMENT]['variant'] = variant

        with open(USER_JSON_PATH, 'w+') as f:
            f.write(json.dumps(old_user_json))



#!/usr/bin/env python
# coding: utf-8

# Copyright (c) Saga Inc.
# Distributed under the terms of the GPL License.
"""
Information about the frontend package of the widgets.
"""

from pathlib import Path
import os
import json

from mitosheet._version import __version__, get_package_json

module_name = get_package_json()['name']
module_version = '^' + __version__



#!/usr/bin/env python
# coding: utf-8

# Copyright (c) Saga Inc.
# Distributed under the terms of the GPL License.
"""
For going to a boolean series.
"""

from typing import Any

import numpy as np
import pandas as pd
from mitosheet.sheet_functions.types.utils import (is_bool_dtype,
                                                   is_datetime_dtype,
                                                   is_number_dtype,
                                                   is_string_dtype)


def to_boolean_series(
        unknown_object: Any,
        on_uncastable_arg_element: Any=('default', np.NaN), # Union[Literal['error'], Tuple[Literal['default'], any]]
    ) -> pd.Series:
    """
    Converts the given object to a boolean series. Note that on_uncastable_arg_element
    is irrelevant here, as anything can be turned into a boolean. 
    """

    # If it is not a series, we put it in a series, and get the type again
    if not isinstance(unknown_object, pd.Series):
        unknown_object = pd.Series([unknown_object])

    column_dtype = str(unknown_object.dtype)
    if is_bool_dtype(column_dtype):
        return unknown_object
    elif is_datetime_dtype(column_dtype):
        # For now, we treat all dates as true, and NaN values as False
        return ~unknown_object.isna()
    elif is_number_dtype(column_dtype):
        return unknown_object.fillna(False).astype('bool')
    elif is_string_dtype(column_dtype):
        string_to_bool_conversion_dict = {
            '1': True,
            '1.0': True,
            1: True,
            1.0: True,
            'TRUE': True,
            'True': True, 
            'true': True,
            'T': True,
            't': True,
            'Y': True,
            'y': True,
            'Yes': True,
            'yes': True,
            #########################
            '0': False,
            '0.0': False,
            0: False,
            0.0: False,
            'FALSE': False,
            'False': False,
            'false': False,
            'F': False,
            'f': False,
            'N': False,
            'n': False,
            'No': False,
            'no': False, 
            'none': False,
            'None': False
        }
        # Convert any NaN values to False
        return unknown_object.map(string_to_bool_conversion_dict).fillna(False)
    else:
        return None



#!/usr/bin/env python
# coding: utf-8

# Copyright (c) Saga Inc.
# Distributed under the terms of the GPL License.
import json
from typing import Any, Dict

import pandas as pd
from mitosheet.sheet_functions.types.utils import is_number_dtype
from mitosheet.types import StepsManagerType


def get_column_describe(params: Dict[str, Any], steps_manager: StepsManagerType) -> str:
    """
    Sends back a string that can be parsed to a JSON object that
    contains _all_ the results from the series .describe function
    for the series at column_header in the df at sheet_index.
    """
    sheet_index = params['sheet_index']
    column_id = params['column_id']
    column_header = steps_manager.curr_step.get_column_header_by_id(sheet_index, column_id)
    
    series: pd.Series = steps_manager.dfs[sheet_index][column_header]
    column_dtype = str(series.dtype)
    describe = series.describe()

    describe_obj = {}

    for index, row in describe.iteritems():
        # We turn all the items to strings, as some items are not valid JSON
        # e.g. some wacky numpy datatypes. This allows us to send all of this 
        # to the front-end.

        # If the series is a number, round the statistics so they look good.
        if is_number_dtype(column_dtype):
            row = round(row, 2)

        describe_obj[index] = str(row)

    # We fill in some specific values that dont get filled by default
    describe_obj['count: NaN'] = str(series.isna().sum())

    # NOTE: be careful adding things here, as we dont want to destroy performance 
    if is_number_dtype(column_dtype):
        describe_obj['median'] = str(round(series.median(), 2))
        describe_obj['sum'] = str(round(series.sum(), 2))

    return json.dumps(describe_obj)
